from unittest import mock

import pytest

from cognite.client.data_classes import Sequence, SequenceColumnUpdate, SequenceFilter, SequenceUpdate
from tests.utils import set_request_limit


@pytest.fixture
def new_seq(cognite_client):
    column_def = [
        {"valueType": "STRING", "externalId": "user", "description": "some description"},
        {"valueType": "DOUBLE", "externalId": "amount"},
        {"valueType": "LONG", "externalId": "age"},
    ]
    seq = cognite_client.sequences.create(Sequence(name="test_temp", columns=column_def, metadata={"a": "b"}))
    yield seq
    cognite_client.sequences.delete(id=seq.id)
    assert cognite_client.sequences.retrieve(seq.id) is None


@pytest.fixture
def get_spy(cognite_client):
    with mock.patch.object(cognite_client.sequences, "_get", wraps=cognite_client.sequences._get) as _:
        yield


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.sequences, "_post", wraps=cognite_client.sequences._post) as _:
        yield


class TestSequencesAPI:
    def test_retrieve(self, cognite_client):
        listed_asset = cognite_client.sequences.list(limit=1)[0]
        retrieved_asset = cognite_client.sequences.retrieve(listed_asset.id)
        assert retrieved_asset == listed_asset

    def test_retrieve_multiple(self, cognite_client):
        res = cognite_client.sequences.list(limit=2)
        retrieved_assets = cognite_client.sequences.retrieve_multiple([s.id for s in res])
        for listed_asset, retrieved_asset in zip(res, retrieved_assets):
            retrieved_asset.external_id = listed_asset.external_id
        assert res == retrieved_assets

    def test_call(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.sequences, 10):
            res = [s for s in cognite_client.sequences(limit=20)]

        assert 20 == len(res)
        assert 2 == cognite_client.sequences._post.call_count

    def test_list(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.sequences, 10):
            res = cognite_client.sequences.list(limit=20)

        assert 20 == len(res)
        assert 2 == cognite_client.sequences._post.call_count

    def test_list_assetid_nothing(self, cognite_client):
        res = cognite_client.sequences.list(asset_ids=[12345678910], limit=20)
        assert 0 == len(res)

    def test_aggregate(self, cognite_client):
        res = cognite_client.sequences.aggregate(filter=SequenceFilter(name="42"))
        assert res[0].count > 0

    def test_search(self, cognite_client):
        res = cognite_client.sequences.search(name="42", filter=SequenceFilter(created_time={"min": 0}))
        assert len(res) > 0

    def test_update(self, cognite_client, new_seq):
        assert new_seq.metadata == {"a": "b"}
        update_seq = SequenceUpdate(new_seq.id).name.set("newname").metadata.set(None)
        res = cognite_client.sequences.update(update_seq)
        assert "newname" == res.name
        assert res.metadata == {}

    def test_update_full(self, cognite_client, new_seq):
        assert new_seq.metadata == {"a": "b"}
        new_seq.name = "newname"
        res = cognite_client.sequences.update(new_seq)
        assert "newname" == res.name

    def test_update_columns_add_remove_single(self, cognite_client, new_seq):
        assert len(new_seq.columns) == 3
        update_seq = SequenceUpdate(new_seq.id).columns.add(
            {"valueType": "STRING", "externalId": "user_added", "description": "some description"}
        )
        res = cognite_client.sequences.update(update_seq)
        assert len(res.columns) == 4
        assert res.column_external_ids[3] == "user_added"

    def test_update_columns_add_multiple(self, cognite_client, new_seq):
        assert len(new_seq.columns) == 3
        column_def = [
            {"valueType": "STRING", "externalId": "user_added", "description": "some description"},
            {"valueType": "DOUBLE", "externalId": "amount_added"},
        ]
        update_seq = SequenceUpdate(new_seq.id).columns.add(column_def)
        res = cognite_client.sequences.update(update_seq)
        assert len(res.columns) == 5
        assert res.column_external_ids[3:5] == ["user_added", "amount_added"]

    def test_update_columns_remove_single(self, cognite_client, new_seq):
        assert len(new_seq.columns) == 3
        update_seq = SequenceUpdate(new_seq.id).columns.remove(new_seq.columns[0]["externalId"])
        res = cognite_client.sequences.update(update_seq)
        assert len(res.columns) == 2
        assert res.columns[0:2] == new_seq.columns[1:3]

    def test_update_columns_remove_multiple(self, cognite_client, new_seq):
        assert len(new_seq.columns) == 3
        update_seq = SequenceUpdate(new_seq.id).columns.remove([col["externalId"] for col in new_seq.columns[0:2]])
        res = cognite_client.sequences.update(update_seq)
        assert len(res.columns) == 1
        assert res.columns[0] == new_seq.columns[2]

    def test_update_columns_modify(self, cognite_client, new_seq):
        assert new_seq.columns[1].get("description") is None
        column_update = [
            SequenceColumnUpdate(external_id=new_seq.columns[0]["externalId"]).external_id.set("new_col_external_id"),
            SequenceColumnUpdate(external_id=new_seq.columns[1]["externalId"]).description.set("my new description"),
        ]
        update_seq = SequenceUpdate(new_seq.id).columns.modify(column_update)
        res = cognite_client.sequences.update(update_seq)
        assert len(res.columns) == 3
        assert res.columns[0]["externalId"] == "new_col_external_id"
        assert res.columns[1]["description"] == "my new description"

    def test_get_new(self, cognite_client, new_seq):
        res = cognite_client.sequences.retrieve(id=new_seq.id)
        # assert ["DOUBLE"] == res.column_value_types # soon to change
        assert len(new_seq.columns) == 3



import inspect
import json
import re
from collections import defaultdict
from doctest import DocTestParser, Example

from cognite.client._api_client import APIClient
from cognite.client.beta import CogniteClient


def collect_apis(obj, done):
    if done.get(obj.__class__):
        return []
    done[obj.__class__] = True
    apis = inspect.getmembers(obj, lambda m: isinstance(m, APIClient) and not done.get(m.__class__))
    sub = [(n + "." + sn, sa) for n, c in apis for sn, sa in collect_apis(c, done)]
    return apis + sub


client = CogniteClient(project="_", api_key="_", client_name="_")
parser = DocTestParser()

apis = collect_apis(client, {})

snippets = {"language": "Python", "label": "Python SDK", "operations": defaultdict(str)}
filter_out = ["from cognite.client import CogniteClient", "c = CogniteClient()", ""]

duplicate_operations = {
    "listAssets": "getAssets",
    "advancedListEvents": "listEvents",
    "advancedListFiles": "listFiles",
    "advancedListSequences": "listSequences",
    "listTimeSeries": "getTimeSeries",
}

for api_name, api in apis:
    for fun_name, fun in inspect.getmembers(api, predicate=inspect.ismethod):
        docstring = fun.__doc__ or ""
        match_link_openapi = re.match("`.* <.*?#operation/(.*)>`_", docstring.strip().split("\n")[0])
        if api_name[0] != "_" and fun_name[0] != "_" and match_link_openapi:
            openapi_ident = match_link_openapi[1]
            parsed_lines = parser.parse(fun.__doc__)
            endpoint_snippets = []
            current_snippet = ""
            for ex in parsed_lines + ["<end>"]:
                if isinstance(ex, Example):
                    if ex.source.strip() not in filter_out:
                        current_snippet += re.sub("(= |in |^)c.", "\\1client.", ex.source.rstrip()) + "\n"
                elif ex != "":
                    if current_snippet:
                        endpoint_snippets.append(current_snippet)
                        current_snippet = ""

            code = "\n".join(endpoint_snippets)
            snippets["operations"][openapi_ident] += code
            if openapi_ident in duplicate_operations:
                snippets["operations"][duplicate_operations[openapi_ident]] += code

print(json.dumps(snippets, indent=2))



import uuid

import pytest

from cognite.client.data_classes import (
    Asset,
    Label,
    LabelDefinition,
    LabelFilter,
    Relationship,
    RelationshipList,
    RelationshipUpdate,
    TimeSeries,
)
from cognite.client.exceptions import CogniteNotFoundError


@pytest.fixture
def new_relationship(new_label, cognite_client):
    external_id = uuid.uuid4().hex[0:20]

    pre_existing_data_set = cognite_client.data_sets.retrieve(external_id="pre_existing_data_set")
    relationship = cognite_client.relationships.create(
        Relationship(
            external_id=external_id,
            source_external_id=external_id,
            source_type="asset",
            target_external_id=external_id,
            data_set_id=pre_existing_data_set.id,
            target_type="event",
            labels=[Label(new_label.external_id)],
            confidence=1,
        )
    )
    yield relationship, external_id
    cognite_client.relationships.delete(external_id=external_id)
    assert cognite_client.relationships.retrieve(external_id=relationship.external_id) is None


@pytest.fixture
def new_label(cognite_client):
    # create a label to use in relationships
    external_id = uuid.uuid4().hex[0:20]
    tp = cognite_client.labels.create(LabelDefinition(external_id=external_id, name="mandatory"))
    assert isinstance(tp, LabelDefinition)
    yield tp
    cognite_client.labels.delete(external_id=tp.external_id)


@pytest.fixture
def new_asset(cognite_client):
    # create an asset to use in relationships
    external_id = uuid.uuid4().hex[0:20]
    tp = cognite_client.assets.create(Asset(external_id=external_id, name="mandatory"))
    assert isinstance(tp, Asset)
    yield tp
    cognite_client.assets.delete(external_id=tp.external_id)


@pytest.fixture
def new_time_series(cognite_client):
    # create a time series to use in relationships
    external_id = uuid.uuid4().hex[0:20]
    tp = cognite_client.time_series.create(TimeSeries(external_id=external_id, name="mandatory"))
    assert isinstance(tp, TimeSeries)
    yield tp
    cognite_client.time_series.delete(external_id=tp.external_id)


@pytest.fixture
def create_multiple_relationships(new_label, cognite_client):
    ext_id = new_label.external_id
    relationships_ext_id = [uuid.uuid4().hex[0:20] for i in range(5)]
    random_ext_id = [uuid.uuid4().hex[0:20] for i in range(5)]
    relationship_list = [
        Relationship(
            external_id=relationships_ext_id[0],
            source_type="asset",
            source_external_id=random_ext_id[0],
            target_type="event",
            target_external_id=random_ext_id[1],
            labels=[Label(ext_id)],
        ),
        Relationship(
            external_id=relationships_ext_id[1],
            source_type="timeSeries",
            source_external_id=random_ext_id[3],
            target_type="asset",
            target_external_id=random_ext_id[2],
            confidence=1,
        ),
        Relationship(
            external_id=relationships_ext_id[2],
            source_type="event",
            source_external_id=random_ext_id[2],
            target_type="timeseries",
            target_external_id=random_ext_id[3],
            labels=[Label(ext_id)],
        ),
        Relationship(
            external_id=relationships_ext_id[3],
            source_type="sequence",
            source_external_id=random_ext_id[3],
            target_type="sequence",
            target_external_id=random_ext_id[4],
            labels=[Label(ext_id)],
        ),
        Relationship(
            external_id=relationships_ext_id[4],
            source_type="file",
            source_external_id=random_ext_id[4],
            target_type="file",
            target_external_id=random_ext_id[0],
            confidence=1,
        ),
    ]
    relationships = cognite_client.relationships.create(relationship_list)
    assert isinstance(relationships, RelationshipList)
    yield relationships_ext_id, ext_id, random_ext_id
    cognite_client.relationships.delete(external_id=[ext_ids["external_id"] for ext_ids in relationships.dump()])


@pytest.fixture
def relationship_with_resources(new_asset, new_time_series, cognite_client):
    external_id = uuid.uuid4().hex[0:20]
    asset_ext_id = new_asset.external_id
    time_series_ext_id = new_time_series.external_id
    relationship = cognite_client.relationships.create(
        Relationship(
            external_id=external_id,
            source_external_id=asset_ext_id,
            source_type="asset",
            target_external_id=time_series_ext_id,
            target_type="timeseries",
        )
    )
    yield relationship, external_id, new_asset, new_time_series
    cognite_client.relationships.delete(external_id=external_id)
    assert cognite_client.relationships.retrieve(external_id=relationship.external_id) is None


class TestRelationshipscognite_client:
    def test_get_single(self, cognite_client, new_relationship):
        new_rel, ext_id = new_relationship
        res = cognite_client.relationships.retrieve(external_id=new_rel.external_id)
        assert isinstance(res, Relationship)
        assert new_rel.external_id == ext_id
        assert res.confidence == 1

    def test_retrieve_unknown(self, cognite_client, new_relationship):
        with pytest.raises(CogniteNotFoundError):
            cognite_client.relationships.retrieve_multiple(external_ids=["this does not exist"])
        assert cognite_client.relationships.retrieve(external_id="this does not exist") is None

    def test_update_single(self, cognite_client, new_relationship):
        new_rel, ext_id = new_relationship
        updated_rel = cognite_client.relationships.update(
            RelationshipUpdate(ext_id).target_type.set("timeseries").confidence.set(None)
        )
        assert isinstance(updated_rel, Relationship)
        assert updated_rel.target_type == "timeSeries"
        assert updated_rel.confidence == None

    def test_list_filter(self, cognite_client, create_multiple_relationships):
        relationships_ext_ids, ext_id, source_ext_id = create_multiple_relationships
        res = cognite_client.relationships.list(source_external_ids=[source_ext_id[3]])
        assert len(res) == 2
        assert isinstance(res, RelationshipList)

    def test_list_data_set(self, cognite_client, new_relationship):
        new_rel, ext_id = new_relationship
        pre_existing_data_set = cognite_client.data_sets.retrieve(external_id="pre_existing_data_set")
        res = cognite_client.relationships.list(
            source_external_ids=[ext_id], data_set_external_ids=[pre_existing_data_set.external_id]
        )
        res2 = cognite_client.relationships.list(target_external_ids=[ext_id], data_set_ids=[pre_existing_data_set.id])
        assert res == res2
        assert isinstance(res, RelationshipList)

    def test_list_label_filter(self, cognite_client, create_multiple_relationships):
        relationships, ext_id, source_ext_id = create_multiple_relationships
        res = cognite_client.relationships.list(labels=LabelFilter(contains_all=[ext_id]))
        assert len(res) == 3
        assert isinstance(res, RelationshipList)

    def test_fetch_resources_list(self, cognite_client, relationship_with_resources):
        relationship, ext_id, asset, time_series = relationship_with_resources
        res = cognite_client.relationships.list(
            source_external_ids=[relationship.source_external_id], fetch_resources=True
        )
        assert res[0].source == asset
        assert res[0].target == time_series

    def test_fetch_resources_retrieve(self, cognite_client, relationship_with_resources):
        relationship, ext_id, asset, time_series = relationship_with_resources
        res = cognite_client.relationships.retrieve_multiple(external_ids=[ext_id], fetch_resources=True)
        assert res[0].source == asset
        assert res[0].target == time_series

    def test_deletes_ignore_unknown_ids(self, cognite_client):
        cognite_client.relationships.delete(external_id=["non_existing_rel"], ignore_unknown_ids=True)

    def test_partitioned_list(self, cognite_client, create_multiple_relationships):
        _, _, ext_ids = create_multiple_relationships
        res_flat = cognite_client.relationships.list(limit=None, source_external_ids=ext_ids)
        res_part = cognite_client.relationships.list(partitions=8, limit=None, source_external_ids=ext_ids)
        assert len(res_flat) > 0
        assert len(res_flat) == len(res_part)
        assert {a.external_id for a in res_flat} == {a.external_id for a in res_part}

    def test_compare_partitioned_gen_and_list(self, cognite_client, create_multiple_relationships):
        _, _, ext_ids = create_multiple_relationships
        res_generator = cognite_client.relationships(partitions=8, limit=None, source_external_ids=ext_ids)
        res_list = cognite_client.relationships.list(partitions=8, limit=None, source_external_ids=ext_ids)
        assert {a.external_id for a in res_generator} == {a.external_id for a in res_list}



import pytest

from cognite.client.data_classes import ExtractionPipeline, ExtractionPipelineRun, ExtractionPipelineUpdate
from cognite.client.data_classes.extractionpipelines import ExtractionPipelineContact
from cognite.client.exceptions import CogniteNotFoundError
from cognite.client.utils._auxiliary import random_string


@pytest.fixture
def new_extpipe(cognite_client):
    testid = random_string(50)
    dataset = cognite_client.data_sets.list()[0]
    extpipe = cognite_client.extraction_pipelines.create(
        ExtractionPipeline(
            external_id=f"testid-{testid}",
            name=f"Test extpipe {testid}",
            data_set_id=dataset.id,
            description="Short description",
            contacts=[
                ExtractionPipelineContact(
                    name="John Doe", email="john.doe@cognite.com", role="owner", send_notification=False
                )
            ],
            schedule="Continuous",
        )
    )
    yield extpipe
    try:
        cognite_client.extraction_pipelines.delete(id=extpipe.id)
    except:
        pass
    assert cognite_client.extraction_pipelines.retrieve(extpipe.id) is None


class TestExtractionPipelinesAPI:
    def test_retrieve(self, cognite_client):
        res = cognite_client.extraction_pipelines.list(limit=1)
        assert res[0] == cognite_client.extraction_pipelines.retrieve(id=res[0].id)

    def test_retrieve_multiple(self, cognite_client):
        res_listed_ids = [e.id for e in cognite_client.extraction_pipelines.list(limit=2)]
        res_lookup_ids = [e.id for e in cognite_client.extraction_pipelines.retrieve_multiple(res_listed_ids)]
        for listed_id in res_listed_ids:
            assert listed_id in res_lookup_ids

    def test_retrieve_unknown(self, cognite_client):
        res = cognite_client.extraction_pipelines.list(limit=1)
        with pytest.raises(CogniteNotFoundError):
            cognite_client.extraction_pipelines.retrieve_multiple(ids=[res[0].id], external_ids=["this does not exist"])
        retr = cognite_client.extraction_pipelines.retrieve_multiple(
            ids=[res[0].id], external_ids=["this does not exist"], ignore_unknown_ids=True
        )
        assert 1 == len(retr)

    def test_update(self, cognite_client, new_extpipe: ExtractionPipeline):
        update_extraction_pipeline = ExtractionPipelineUpdate(id=new_extpipe.id).metadata.set({"hey": "now"})
        res = cognite_client.extraction_pipelines.update(update_extraction_pipeline)
        assert {"hey": "now"} == res.metadata
        update_extraction_pipeline2 = ExtractionPipelineUpdate(id=new_extpipe.id).metadata.set(None)
        res2 = cognite_client.extraction_pipelines.update(update_extraction_pipeline2)
        assert res2.metadata is None

    def test_delete(self, cognite_client, new_extpipe: ExtractionPipeline):
        cognite_client.extraction_pipelines.delete(id=new_extpipe.id)
        assert cognite_client.extraction_pipelines.retrieve(id=new_extpipe.id) is None

    def test_last_status_message(self, cognite_client, new_extpipe: ExtractionPipeline):
        intial_message = cognite_client.extraction_pipelines.retrieve(new_extpipe.id).last_message
        assert intial_message is None

        cognite_client.extraction_pipeline_runs.create(
            ExtractionPipelineRun(external_id=new_extpipe.external_id, status="failure", message="Oh no!")
        )
        new_message = cognite_client.extraction_pipelines.retrieve(new_extpipe.id).last_message
        assert new_message == "Oh no!"

    def test_last_run_times(self, cognite_client, new_extpipe: ExtractionPipeline):
        extpipe = cognite_client.extraction_pipelines.retrieve(external_id=new_extpipe.external_id)
        assert extpipe.last_failure == 0
        assert extpipe.last_success == 0
        assert extpipe.last_seen == 0

        cognite_client.extraction_pipeline_runs.create(
            ExtractionPipelineRun(external_id=new_extpipe.external_id, status="seen")
        )
        extpipe = cognite_client.extraction_pipelines.retrieve(external_id=new_extpipe.external_id)
        assert extpipe.last_failure == 0
        assert extpipe.last_success == 0
        assert extpipe.last_seen != 0

        cognite_client.extraction_pipeline_runs.create(
            ExtractionPipelineRun(external_id=new_extpipe.external_id, status="success")
        )
        extpipe = cognite_client.extraction_pipelines.retrieve(external_id=new_extpipe.external_id)
        assert extpipe.last_failure == 0
        assert extpipe.last_success != 0
        assert extpipe.last_seen != 0

        cognite_client.extraction_pipeline_runs.create(
            ExtractionPipelineRun(external_id=new_extpipe.external_id, status="failure")
        )
        extpipe = cognite_client.extraction_pipelines.retrieve(external_id=new_extpipe.external_id)
        assert extpipe.last_failure != 0
        assert extpipe.last_success != 0
        assert extpipe.last_seen != 0



from unittest import mock

import pytest

from cognite.client.data_classes import DataSet, DataSetFilter, DataSetUpdate
from cognite.client.exceptions import CogniteNotFoundError
from tests.utils import set_request_limit


@pytest.fixture(scope="class")
def new_dataset(cognite_client):
    dataset = cognite_client.data_sets.create(DataSet())
    yield dataset
    # todo: uncomment when delete is implemented
    # cognite_client.data_sets.delete(id=dataset.id)
    # assert cognite_client.data_sets.retrieve(dataset.id) is None


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.data_sets, "_post", wraps=cognite_client.data_sets._post) as _:
        yield


class TestDataSetsAPI:
    def test_retrieve(self, cognite_client, new_dataset):
        assert new_dataset.id == cognite_client.data_sets.retrieve(new_dataset.id).id

    def test_retrieve_multiple(self, cognite_client):
        res_listed_ids = [e.id for e in cognite_client.data_sets.list(limit=2)]
        res_lookup_ids = [e.id for e in cognite_client.data_sets.retrieve_multiple(res_listed_ids)]
        for listed_id in res_listed_ids:
            assert listed_id in res_lookup_ids

    def test_retrieve_unknown(self, cognite_client, new_dataset):
        invalid_external_id = "this does not exist"
        with pytest.raises(CogniteNotFoundError) as error:
            cognite_client.data_sets.retrieve_multiple(ids=[new_dataset.id], external_ids=[invalid_external_id])
        assert error.value.not_found[0]["externalId"] == invalid_external_id

    def test_list(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.data_sets, 1):
            res = cognite_client.data_sets.list(limit=2)

        assert 2 == len(res)
        assert 2 == cognite_client.data_sets._post.call_count

    def test_aggregate(self, cognite_client):
        res = cognite_client.data_sets.aggregate(filter=DataSetFilter(metadata={"1": "1"}))
        assert res[0].count > 0

    def test_update(self, cognite_client, new_dataset):
        update_asset = DataSetUpdate(new_dataset.id).metadata.set({"1": "1"}).name.set("newname")
        res = cognite_client.data_sets.update(update_asset)
        assert {"1": "1"} == res.metadata
        assert "newname" == res.name



from unittest import mock

import pytest

from cognite.client.data_classes import Asset, AssetUpdate, Label, LabelDefinition
from cognite.client.utils._auxiliary import random_string


@pytest.fixture
def new_label(cognite_client):
    tp = cognite_client.labels.create(LabelDefinition(external_id=random_string(30), name="mandatory"))
    yield tp
    cognite_client.labels.delete(external_id=tp.external_id)


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.labels, "_post", wraps=cognite_client.labels._post) as _:
        yield


class TestLabelsAPI:
    def test_list(self, cognite_client, new_label, post_spy):
        res = cognite_client.labels.list(limit=100)
        assert 0 < len(res) <= 100
        assert 1 == cognite_client.labels._post.call_count

    def test_create_asset_with_label(self, cognite_client, new_label):
        ac = cognite_client.assets.create(Asset(name="any", labels=[Label(external_id=new_label.external_id)]))
        assert isinstance(ac, Asset)
        assert len(ac.labels) == 1
        cognite_client.assets.delete(id=ac.id)

    def test_update_asset_with_label(self, cognite_client, new_label):
        ac = cognite_client.assets.create(Asset(name="any", description="delete me"))
        assert not ac.labels
        update = AssetUpdate(id=ac.id).labels.add([new_label.external_id])
        assert isinstance(update, AssetUpdate)
        ua = cognite_client.assets.update(update)
        assert len(ua.labels) == 1
        assert new_label.external_id == ua.labels[0].external_id

        update = AssetUpdate(id=ac.id).labels.remove([new_label.external_id])
        assert isinstance(update, AssetUpdate)
        ua = cognite_client.assets.update(update)

        assert not ua.labels
        cognite_client.assets.delete(id=ac.id)



from decimal import Decimal
from unittest import mock

import pytest

from cognite.client import CogniteClient
from cognite.client.data_classes._base import *
from cognite.client.exceptions import CogniteMissingClientError
from tests.utils import set_env_var


class MyResource(CogniteResource):
    def __init__(self, var_a=None, var_b=None, id=None, external_id=None, cognite_client=None):
        self.var_a = var_a
        self.var_b = var_b
        self.id = id
        self.external_id = external_id
        self._cognite_client = cognite_client

    def use(self):
        return self._cognite_client


class MyUpdate(CogniteUpdate):
    @property
    def string(self):
        return PrimitiveUpdate(self, "string")

    @property
    def list(self):
        return ListUpdate(self, "list")

    @property
    def object(self):
        return ObjectUpdate(self, "object")

    @property
    def labels(self):
        return LabelUpdate(self, "labels")


class PrimitiveUpdate(CognitePrimitiveUpdate):
    def set(self, value: Any) -> MyUpdate:
        return self._set(value)


class ObjectUpdate(CogniteObjectUpdate):
    def set(self, value: Dict) -> MyUpdate:
        return self._set(value)

    def add(self, value: Dict) -> MyUpdate:
        return self._add(value)

    def remove(self, value: List) -> MyUpdate:
        return self._remove(value)


class ListUpdate(CogniteListUpdate):
    def set(self, value: List) -> MyUpdate:
        return self._set(value)

    def add(self, value: List) -> MyUpdate:
        return self._add(value)

    def remove(self, value: List) -> MyUpdate:
        return self._remove(value)


class LabelUpdate(CogniteLabelUpdate):
    def add(self, value: List) -> MyUpdate:
        return self._add(value)

    def remove(self, value: List) -> MyUpdate:
        return self._remove(value)


class MyFilter(CogniteFilter):
    def __init__(self, var_a=None, var_b=None, cognite_client=None):
        self.var_a = var_a
        self.var_b = var_b
        self._cognite_client = cognite_client

    def use(self):
        return self._cognite_client


class MyResourceList(CogniteResourceList):
    _RESOURCE = MyResource
    _UPDATE = MyUpdate

    def use(self):
        return self._cognite_client


class MyResponse(CogniteResponse):
    def __init__(self, var_a=None, cognite_client=None):
        self.var_a = var_a
        self._cognite_client = cognite_client

    @classmethod
    def _load(cls, api_response):
        data = api_response["data"]
        return cls(data["varA"])

    def use(self):
        return self._cognite_client


class TestCogniteResource:
    def test_dump(self):
        assert {"var_a": 1} == MyResource(1).dump()
        assert {"var_a": 1} == MyResource(1).dump(camel_case=False)

    def test_dump_camel_case(self):
        assert {"varA": 1} == MyResource(1).dump(camel_case=True)

    def test_load(self):
        assert MyResource(1).dump() == MyResource._load({"varA": 1}).dump()
        assert MyResource(1, 2).dump() == MyResource._load({"var_a": 1, "var_b": 2}).dump()
        assert {"var_a": 1} == MyResource._load({"var_a": 1, "var_c": 1}).dump()

    def test_load_unknown_attribute(self):
        assert {"var_a": 1, "var_b": 2} == MyResource._load({"varA": 1, "varB": 2, "varC": 3}).dump()

    def test_load_object_attr(self):
        assert {"var_a": 1, "var_b": {"camelCase": 1}} == MyResource._load({"varA": 1, "varB": {"camelCase": 1}}).dump()

    def test_eq(self):
        assert MyResource(1, "s") == MyResource(1, "s")
        assert MyResource(1, "s") == MyResource(1, "s", cognite_client=mock.MagicMock())
        assert MyResource() == MyResource()
        assert MyResource(1, "s") != MyResource(1)
        assert MyResource(1, "s") != MyResource(2, "t")

    def test_str_repr(self):
        assert json.dumps({"var_a": 1}, indent=4) == MyResource(1).__str__()
        assert json.dumps({"var_a": 1.0}, indent=4) == MyResource(Decimal(1)).__str__()

    @pytest.mark.dsl
    def test_to_pandas(self):
        import pandas as pd

        class SomeResource(CogniteResource):
            def __init__(self, a_list, ob, ob_expand, ob_ignore, prim, prim_ignore):
                self.a_list = a_list
                self.ob = ob
                self.ob_expand = ob_expand
                self.ob_ignore = ob_ignore
                self.prim = prim
                self.prim_ignore = prim_ignore

        expected_df = pd.DataFrame(columns=["value"])
        expected_df.loc["prim"] = ["abc"]
        expected_df.loc["aList"] = [[1, 2, 3]]
        expected_df.loc["ob"] = [{"x": "y"}]
        expected_df.loc["md_key"] = ["md_value"]

        res = SomeResource([1, 2, 3], {"x": "y"}, {"md_key": "md_value"}, {"bla": "bla"}, "abc", 1)
        actual_df = res.to_pandas(expand=["obExpand"], ignore=["primIgnore", "obIgnore"])
        pd.testing.assert_frame_equal(expected_df, actual_df, check_like=True)
        res.to_pandas()

    @pytest.mark.dsl
    def test_to_pandas_no_camels(self):
        import pandas as pd

        class SomeResource(CogniteResource):
            def __init__(self):
                self.snakes_are_better_anyway = 42

        expected_df = pd.DataFrame(columns=["value"])
        expected_df.loc["snakes_are_better_anyway"] = [42]

        actual_df = SomeResource().to_pandas(camel_case=False)
        pd.testing.assert_frame_equal(expected_df, actual_df, check_like=True)

    def test_resource_client_correct(self):
        with set_env_var("COGNITE_API_KEY", "BLA"):
            c = CogniteClient()
        with pytest.raises(CogniteMissingClientError):
            MyResource(1)._cognite_client
        assert MyResource(1, cognite_client=c)._cognite_client == c

    def test_use_method_which_requires_cognite_client__client_not_set(self):
        mr = MyResource()
        with pytest.raises(CogniteMissingClientError):
            mr.use()


class TestCogniteResourceList:
    def test_dump(self):
        assert [{"var_a": 1, "var_b": 2}, {"var_a": 2, "var_b": 3}] == MyResourceList(
            [MyResource(1, 2), MyResource(2, 3)]
        ).dump()

    @pytest.mark.dsl
    def test_to_pandas(self):
        import pandas as pd

        resource_list = MyResourceList([MyResource(1), MyResource(2, 3)])
        expected_df = pd.DataFrame({"varA": [1, 2], "varB": [None, 3]})
        pd.testing.assert_frame_equal(resource_list.to_pandas(), expected_df)

    @pytest.mark.dsl
    def test_to_pandas_no_camels(self):
        import pandas as pd

        resource_list = MyResourceList([MyResource(1), MyResource(2, 3)])
        expected_df = pd.DataFrame({"var_a": [1, 2], "var_b": [None, 3]})
        pd.testing.assert_frame_equal(resource_list.to_pandas(camel_case=False), expected_df)

    def test_load(self):
        resource_list = MyResourceList._load([{"varA": 1, "varB": 2}, {"varA": 2, "varB": 3}, {"varA": 3}])

        assert {"var_a": 1, "var_b": 2} == resource_list[0].dump()
        assert [{"var_a": 1, "var_b": 2}, {"var_a": 2, "var_b": 3}, {"var_a": 3}] == resource_list.dump()

    def test_load_unknown_attribute(self):
        assert [{"var_a": 1, "var_b": 2}] == MyResourceList._load([{"varA": 1, "varB": 2, "varC": 3}]).dump()

    def test_indexing(self):
        resource_list = MyResourceList([MyResource(1, 2), MyResource(2, 3)])
        assert MyResource(1, 2) == resource_list[0]
        assert MyResource(2, 3) == resource_list[1]
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3)]) == resource_list[:]
        assert isinstance(resource_list[:], MyResourceList)

    def test_slice_list_client_remains(self):
        mock_client = mock.MagicMock()
        rl = MyResourceList([MyResource(1, 2)], cognite_client=mock_client)
        rl_sliced = rl[:]
        assert rl._cognite_client == rl_sliced._cognite_client

    def test_extend(self):
        resource_list = MyResourceList([MyResource(1, 2), MyResource(2, 3)])
        another_resource_list = MyResourceList([MyResource(4, 5), MyResource(6, 7)])
        resource_list.extend(another_resource_list)
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3), MyResource(4, 5), MyResource(6, 7)]) == resource_list

    def test_len(self):
        resource_list = MyResourceList([MyResource(1, 2), MyResource(2, 3)])
        assert 2 == len(resource_list)

    def test_eq(self):
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3)]) == MyResourceList(
            [MyResource(1, 2), MyResource(2, 3)]
        )
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3)]) != MyResourceList(
            [MyResource(2, 3), MyResource(1, 2)]
        )
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3)]) != MyResourceList(
            [MyResource(2, 3), MyResource(1, 4)]
        )
        assert MyResourceList([MyResource(1, 2), MyResource(2, 3)]) != MyResourceList([MyResource(1, 2)])

    def test_iter(self):
        resource_list = MyResourceList([MyResource(1, 2), MyResource(2, 3)])
        counter = 0
        for resource in resource_list:
            counter += 1
            assert resource in [MyResource(1, 2), MyResource(2, 3)]
        assert 2 == counter

    def test_get_item_by_id(self):
        resource_list = MyResourceList([MyResource(id=1, external_id="1"), MyResource(id=2, external_id="2")])
        assert MyResource(id=1, external_id="1") == resource_list.get(id=1)
        assert MyResource(id=2, external_id="2") == resource_list.get(id=2)

    def test_str_repr(self):
        assert json.dumps([{"var_a": 1}], indent=4) == MyResourceList([MyResource(1)]).__str__()
        assert json.dumps([{"var_a": 1.0}], indent=4) == MyResourceList([MyResource(Decimal(1))]).__str__()

    def test_get_item_by_external_id(self):
        resource_list = MyResourceList([MyResource(id=1, external_id="1"), MyResource(id=2, external_id="2")])
        assert MyResource(id=1, external_id="1") == resource_list.get(external_id="1")
        assert MyResource(id=2, external_id="2") == resource_list.get(external_id="2")

    def test_constructor_bad_type(self):
        with pytest.raises(TypeError, match="must be of type 'MyResource'"):
            MyResourceList([1, 2, 3])

    def test_resource_list_client_correct(self):
        with set_env_var("COGNITE_API_KEY", "bla"):
            c = CogniteClient()
        with pytest.raises(CogniteMissingClientError):
            MyResource(1)._cognite_client
        assert MyResource(1, cognite_client=c)._cognite_client == c

    def test_use_method_which_requires_cognite_client__client_not_set(self):
        mr = MyResourceList([])
        with pytest.raises(CogniteMissingClientError):
            mr.use()


class TestCogniteFilter:
    def test_dump(self):
        assert MyFilter(1, 2).dump() == {"var_a": 1, "var_b": 2}
        assert MyFilter(1, 2).dump(camel_case=True) == {"varA": 1, "varB": 2}

    def test_eq(self):
        assert MyFilter(1, 2) == MyFilter(1, 2)
        assert MyFilter(1, 2) == MyFilter(1, 2, cognite_client=mock.MagicMock())
        assert MyFilter(1) != MyFilter(1, 2)
        assert MyFilter() == MyFilter()

    def test_str(self):
        assert json.dumps({"var_a": 1}, indent=4) == MyFilter(1).__str__()
        assert json.dumps({"var_a": 1.0}, indent=4) == MyFilter(Decimal(1)).__str__()

    def test_repr(self):
        assert json.dumps({"var_a": 1}, indent=4) == MyFilter(1).__repr__()

    def test_use_method_which_requires_cognite_client__client_not_set(self):
        mr = MyFilter()
        with pytest.raises(CogniteMissingClientError):
            mr.use()


class TestCogniteUpdate:
    def test_dump_id(self):
        assert {"id": 1, "update": {}} == MyUpdate(id=1).dump()

    def test_dump_external_id(self):
        assert {"externalId": "1", "update": {}} == MyUpdate(external_id="1").dump()

    def test_dump_both_ids_set(self):
        assert {"id": 1, "update": {}} == MyUpdate(id=1, external_id="1").dump()

    def test_eq(self):
        assert MyUpdate() == MyUpdate()
        assert MyUpdate(1) == MyUpdate(1)
        assert MyUpdate(1).string.set("1") == MyUpdate(1).string.set("1")
        assert MyUpdate(1) != MyUpdate(2)
        assert MyUpdate(1) != MyUpdate(1).string.set("1")

    def test_str(self):
        assert json.dumps(MyUpdate(1).dump(), indent=4) == MyUpdate(1).__str__()
        assert json.dumps(MyUpdate(1.0).dump(), indent=4) == MyUpdate(Decimal(1)).__str__()
        assert json.dumps(MyUpdate(1).string.set("1").dump(), indent=4) == MyUpdate(1).string.set("1").__str__()

    def test_set_string(self):
        assert {"id": 1, "update": {"string": {"set": "bla"}}} == MyUpdate(1).string.set("bla").dump()

    def test_add_to_list(self):
        assert {"id": 1, "update": {"list": {"add": [1, 2, 3]}}} == MyUpdate(1).list.add([1, 2, 3]).dump()

    def test_set_list(self):
        assert {"id": 1, "update": {"list": {"set": [1, 2, 3]}}} == MyUpdate(1).list.set([1, 2, 3]).dump()

    def test_remove_from_list(self):
        assert {"id": 1, "update": {"list": {"remove": [1, 2, 3]}}} == MyUpdate(1).list.remove([1, 2, 3]).dump()

    def test_set_object(self):
        assert {"id": 1, "update": {"object": {"set": {"key": "value"}}}} == MyUpdate(1).object.set(
            {"key": "value"}
        ).dump()

    def test_add_object(self):
        assert {"id": 1, "update": {"object": {"add": {"key": "value"}}}} == MyUpdate(1).object.add(
            {"key": "value"}
        ).dump()

    def test_add_or_remove_after_set_raises_error(self):
        update = MyUpdate(1).object.set({"key": "value"})
        with pytest.raises(AssertionError):
            update.object.add({"key2": "value2"})
        with pytest.raises(AssertionError):
            update.object.remove(["key2"])

    def test_set_after_add_or_removeraises_error(self):
        update = MyUpdate(1).object.add({"key": "value"})
        with pytest.raises(AssertionError):
            update.object.set({"key2": "value2"})

    def test_add_object_and_remove(self):
        update = MyUpdate(1).object.add({"key": "value"})
        update.object.remove(["key2"])
        assert {"id": 1, "update": {"object": {"add": {"key": "value"}, "remove": ["key2"]}}} == update.dump()
        with pytest.raises(AssertionError):
            update.object.add({"key": "overwrite"})
        with pytest.raises(AssertionError):
            update.object.remove(["key2", "key4"])

    def test_remove_object(self):
        assert {"id": 1, "update": {"object": {"remove": ["value"]}}} == MyUpdate(1).object.remove(["value"]).dump()

    def test_set_string_null(self):
        assert {"externalId": "1", "update": {"string": {"setNull": True}}} == MyUpdate(external_id="1").string.set(
            None
        ).dump()

    def test_chain_setters(self):
        assert {"id": 1, "update": {"object": {"set": {"bla": "bla"}}, "string": {"set": "bla"}}} == MyUpdate(
            id=1
        ).object.set({"bla": "bla"}).string.set("bla").dump()

    def test_get_update_properties(self):
        assert {"string", "list", "object"} == set(MyUpdate._get_update_properties())


class TestCogniteResponse:
    def test_load(self):
        res = MyResponse._load({"data": {"varA": 1}})
        assert 1 == res.var_a

    def test_dump(self):
        assert {"var_a": 1} == MyResponse(1).dump()
        assert {"varA": 1} == MyResponse(1).dump(camel_case=True)
        assert {} == MyResponse().dump()

    def test_str(self):
        assert json.dumps(MyResponse(1).dump(), indent=4, sort_keys=True) == MyResponse(1).__str__()
        assert json.dumps(MyResponse(1.0).dump(), indent=4, sort_keys=True) == MyResponse(Decimal(1)).__str__()

    def test_repr(self):
        assert json.dumps(MyResponse(1).dump(), indent=4, sort_keys=True) == MyResponse(1).__repr__()

    def test_eq(self):
        assert MyResponse(1) == MyResponse(1)
        assert MyResponse(1) == MyResponse(1, cognite_client=mock.MagicMock())
        assert MyResponse(1) != MyResponse(2)
        assert MyResponse(1) != MyResponse()

    def test_response_client_correct(self):
        with set_env_var("COGNITE_API_KEY", "BLA"):
            c = CogniteClient()
        with pytest.raises(CogniteMissingClientError):
            MyResource(1)._cognite_client
        assert MyResource(1, cognite_client=c)._cognite_client == c

    def test_use_method_which_requires_cognite_client__client_not_set(self):
        mr = MyResponse()
        with pytest.raises(CogniteMissingClientError):
            mr.use()



import json
import queue
import re
import time
from collections import OrderedDict

import pytest

from cognite.client._api.assets import Asset, AssetList, AssetUpdate, _AssetPoster, _AssetPosterWorker
from cognite.client.data_classes import AssetFilter, Label, LabelFilter, TimestampRange
from cognite.client.exceptions import CogniteAPIError
from tests.utils import jsgz_load, set_request_limit

EXAMPLE_ASSET = {
    "externalId": "string",
    "name": "string",
    "parentId": 1,
    "description": "string",
    "metadata": {"metadata-key": "metadata-value"},
    "labels": [{"externalId": "PUMP"}],
    "source": "string",
    "id": 1,
    "lastUpdatedTime": 0,
    "rootId": 1,
}


@pytest.fixture
def mock_assets_response(rsps, cognite_client):
    response_body = {"items": [EXAMPLE_ASSET]}
    url_pattern = re.compile(re.escape(cognite_client.assets._get_base_url_with_base_path()) + "/.+")
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_get_subtree_base(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/byids",
        status=200,
        json={"items": [{"id": 1}]},
    )
    rsps.add(
        rsps.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=200,
        json={"items": [{"id": 2, "parentId": 1}, {"id": 3, "parentId": 1}, {"id": 4, "parentId": 1}]},
    )
    rsps.add(
        rsps.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=200,
        json={"items": [{"id": 5, "parentId": 2}, {"id": 6, "parentId": 2}]},
    )
    rsps.add(
        rsps.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=200,
        json={"items": [{"id": 7, "parentId": 3}, {"id": 8, "parentId": 3}]},
    )
    rsps.add(
        rsps.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=200,
        json={"items": [{"id": 9, "parentId": 4}, {"id": 10, "parentId": 4}]},
    )
    yield rsps


@pytest.fixture
def mock_get_subtree(mock_get_subtree_base, cognite_client):
    mock_get_subtree_base.add(
        mock_get_subtree_base.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=200,
        json={"items": []},
    )
    yield mock_get_subtree_base


@pytest.fixture
def mock_get_subtree_w_request_failure(mock_get_subtree_base, cognite_client):
    mock_get_subtree_base.add(
        mock_get_subtree_base.POST,
        cognite_client.assets._get_base_url_with_base_path() + "/assets/list",
        status=503,
        json={"error": {"message": "Service Unavailable"}},
    )
    yield mock_get_subtree_base


class TestAssets:
    def test_retrieve_single(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.retrieve(id=1)
        assert isinstance(res, Asset)
        assert mock_assets_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_retrieve_multiple(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.retrieve_multiple(ids=[1])
        assert isinstance(res, AssetList)
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.list(name="bla")
        assert "bla" == jsgz_load(mock_assets_response.calls[0].request.body)["filter"]["name"]
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_with_aggregated_properties_param(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(name="bla", aggregated_properties=["childCount"])
        assert ["childCount"] == jsgz_load(mock_assets_response.calls[0].request.body)["aggregatedProperties"]

    def test_list_with_aggregated_properties_param_when_snake_cased(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(name="bla", aggregated_properties=["child_count"])
        assert ["childCount"] == jsgz_load(mock_assets_response.calls[0].request.body)["aggregatedProperties"]

    def test_list_with_dataset_ids(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(name="bla", data_set_ids=[1], data_set_external_ids=["x"])
        assert [{"id": 1}, {"externalId": "x"}] == jsgz_load(mock_assets_response.calls[0].request.body)["filter"][
            "dataSetIds"
        ]

    def test_list_root(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(root_ids=[{"id": 1}, {"externalId": "abc"}], limit=10)
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {"rootIds": [{"id": 1}, {"externalId": "abc"}]}} == jsgz_load(
            calls[0].request.body
        )

    def test_list_parent(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(parent_ids=[1, 2], parent_external_ids=["abc"], limit=10)
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"parentIds": [1, 2], "parentExternalIds": ["abc"]},
        } == jsgz_load(calls[0].request.body)

    def test_list_subtree(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(asset_subtree_ids=1, asset_subtree_external_ids=["a"], limit=10)
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"assetSubtreeIds": [{"id": 1}, {"externalId": "a"}]},
        } == jsgz_load(calls[0].request.body)

    def test_list_with_time_dict(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(created_time={"min": 20})
        assert 20 == jsgz_load(mock_assets_response.calls[0].request.body)["filter"]["createdTime"]["min"]
        assert "max" not in jsgz_load(mock_assets_response.calls[0].request.body)["filter"]["createdTime"]

    def test_list_with_timestamp_range(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(created_time=TimestampRange(min=20))
        assert 20 == jsgz_load(mock_assets_response.calls[0].request.body)["filter"]["createdTime"]["min"]
        assert "max" not in jsgz_load(mock_assets_response.calls[0].request.body)["filter"]["createdTime"]

    def test_call_root(self, cognite_client, mock_assets_response):
        list(cognite_client.assets.__call__(root_ids=[{"id": 1}, {"externalId": "abc"}], limit=10))
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {"rootIds": [{"id": 1}, {"externalId": "abc"}]}} == jsgz_load(
            calls[0].request.body
        )

    def test_list_root_ids_list(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(root_ids=[1, 2], limit=10)
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {"rootIds": [{"id": 1}, {"id": 2}]}} == jsgz_load(
            calls[0].request.body
        )

    def test_list_root_extids_list(self, cognite_client, mock_assets_response):
        cognite_client.assets.list(root_external_ids=["1", "2"], limit=10)
        calls = mock_assets_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"rootIds": [{"externalId": "1"}, {"externalId": "2"}]},
        } == jsgz_load(calls[0].request.body)

    def test_create_single(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.create(Asset(external_id="1", name="blabla"))
        assert isinstance(res, Asset)
        assert mock_assets_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.create([Asset(external_id="1", name="blabla")])
        assert isinstance(res, AssetList)
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_iter_single(self, cognite_client, mock_assets_response):
        for asset in cognite_client.assets:
            assert mock_assets_response.calls[0].response.json()["items"][0] == asset.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_assets_response):
        for assets in cognite_client.assets(chunk_size=1):
            assert mock_assets_response.calls[0].response.json()["items"] == assets.dump(camel_case=True)

    def test_delete_single(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.delete(id=1)
        assert {"items": [{"id": 1}], "recursive": False, "ignoreUnknownIds": False} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )
        assert res is None

    def test_delete_single_recursive(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.delete(id=1, recursive=True)
        assert {"items": [{"id": 1}], "recursive": True, "ignoreUnknownIds": False} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.delete(id=[1], ignore_unknown_ids=True)
        assert {"items": [{"id": 1}], "recursive": False, "ignoreUnknownIds": True} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.update(Asset(id=1))
        assert isinstance(res, Asset)
        assert mock_assets_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.update(AssetUpdate(id=1).description.set("blabla"))
        assert isinstance(res, Asset)
        assert mock_assets_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.update([AssetUpdate(id=1).description.set("blabla")])
        assert isinstance(res, AssetList)
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_labels_single(self, cognite_client, mock_assets_response):
        cognite_client.assets.update([AssetUpdate(id=1).labels.add("PUMP").labels.remove("VALVE")])
        expected = {"labels": {"add": [{"externalId": "PUMP"}], "remove": [{"externalId": "VALVE"}]}}
        assert expected == jsgz_load(mock_assets_response.calls[0].request.body)["items"][0]["update"]

    def test_update_labels_multiple(self, cognite_client, mock_assets_response):
        cognite_client.assets.update(
            [AssetUpdate(id=1).labels.add(["PUMP", "ROTATING_EQUIPMENT"]).labels.remove(["VALVE", "VERIFIED"])]
        )
        expected = {
            "labels": {
                "add": [{"externalId": "PUMP"}, {"externalId": "ROTATING_EQUIPMENT"}],
                "remove": [{"externalId": "VALVE"}, {"externalId": "VERIFIED"}],
            }
        }
        assert expected == jsgz_load(mock_assets_response.calls[0].request.body)["items"][0]["update"]

    def test_update_labels_set_single(self, cognite_client, mock_assets_response):
        cognite_client.assets.update([AssetUpdate(id=1).labels.set("PUMP")])
        expected = {"labels": {"set": [{"externalId": "PUMP"}]}}
        assert expected == jsgz_load(mock_assets_response.calls[0].request.body)["items"][0]["update"]

    def test_update_labels_set_multiple(self, cognite_client, mock_assets_response):
        cognite_client.assets.update([AssetUpdate(id=1).labels.set(["PUMP", "VALVE"])])
        expected = {"labels": {"set": [{"externalId": "PUMP"}, {"externalId": "VALVE"}]}}
        assert expected == jsgz_load(mock_assets_response.calls[0].request.body)["items"][0]["update"]

    # resource.update doesn't support full replacement of labels (set operation)
    def test_ignore_labels_resource_class(self, cognite_client, mock_assets_response):
        cognite_client.assets.update(Asset(id=1, labels=[Label(external_id="Pump")], name="Abc"))
        assert {"name": {"set": "Abc"}} == jsgz_load(mock_assets_response.calls[0].request.body)["items"][0]["update"]

    def test_labels_filter_contains_all(self, cognite_client, mock_assets_response):
        my_label_filter = LabelFilter(contains_all=["PUMP", "VERIFIED"])
        cognite_client.assets.list(labels=my_label_filter)
        assert {"containsAll": [{"externalId": "PUMP"}, {"externalId": "VERIFIED"}]} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )["filter"]["labels"]

    def test_labels_filter_contains_any(self, cognite_client, mock_assets_response):
        my_label_filter = LabelFilter(contains_any=["PUMP", "VALVE"])
        cognite_client.assets.list(labels=my_label_filter)
        assert {"containsAny": [{"externalId": "PUMP"}, {"externalId": "VALVE"}]} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )["filter"]["labels"]

    def test_create_asset_with_label(self, cognite_client, mock_assets_response):
        cognite_client.assets.create(
            Asset(name="test", labels=[Label(external_id="PUMP"), Label(external_id="VERIFIED")])
        )
        assert {"name": "test", "labels": [{"externalId": "PUMP"}, {"externalId": "VERIFIED"}]} == jsgz_load(
            mock_assets_response.calls[0].request.body
        )["items"][0]

    def test_search(self, cognite_client, mock_assets_response):
        res = cognite_client.assets.search(filter=AssetFilter(name="1"))
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"name": "1"},
            "limit": 100,
        } == jsgz_load(mock_assets_response.calls[0].request.body)

    @pytest.mark.parametrize("filter_field", ["parent_ids", "parentIds"])
    def test_search_dict_filter(self, cognite_client, mock_assets_response, filter_field):
        res = cognite_client.assets.search(filter={filter_field: "bla"})
        assert mock_assets_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {
            "search": {"name": None, "description": None, "query": None},
            "filter": {"parentIds": "bla"},
            "limit": 100,
        } == jsgz_load(mock_assets_response.calls[0].request.body)

    def test_get_subtree(self, cognite_client, mock_get_subtree):
        assets = cognite_client.assets.retrieve_subtree(id=1)
        assert len(assets) == 10
        for i, asset in enumerate(assets):
            assert asset.id == i + 1

    def test_get_subtree_w_depth(self, cognite_client, mock_get_subtree):
        mock_get_subtree.assert_all_requests_are_fired = False
        assets = cognite_client.assets.retrieve_subtree(id=1, depth=1)
        assert len(assets) == 4
        for i, asset in enumerate(assets):
            assert asset.id == i + 1

    def test_get_subtree_w_error(self, cognite_client, mock_get_subtree_w_request_failure):
        with pytest.raises(CogniteAPIError):
            cognite_client.assets.retrieve_subtree(id=1)

    def test_assets_update_object(self):
        assert isinstance(
            AssetUpdate(1)
            .description.set("")
            .description.set(None)
            .external_id.set("1")
            .external_id.set(None)
            .metadata.set({})
            .metadata.set(None)
            .labels.add(["PUMP"])
            .labels.remove(["VALVE"])
            .name.set("")
            .name.set(None)
            .source.set(1)
            .source.set(None)
            .data_set_id.set(123),
            AssetUpdate,
        )


class TestAssetPosterWorker:
    def test_run(self, cognite_client, mock_assets_response):
        q_req = queue.Queue()
        q_res = queue.Queue()

        w = _AssetPosterWorker(request_queue=q_req, response_queue=q_res, client=cognite_client.assets)
        w.start()
        q_req.put([Asset()])
        time.sleep(0.1)
        w.stop = True
        assert [Asset._load(mock_assets_response.calls[0].response.json()["items"][0])] == q_res.get()
        assert 1 == len(mock_assets_response.calls)


def generate_asset_tree(root_external_id: str, depth: int, children_per_node: int, current_depth=1):
    assert 1 <= children_per_node <= 10, "children_per_node must be between 1 and 10"
    assets = []
    if current_depth == 1:
        assets = [Asset(external_id=root_external_id)]
    if depth > current_depth:
        for i in range(children_per_node):
            asset = Asset(parent_external_id=root_external_id, external_id="{}{}".format(root_external_id, i))
            assets.append(asset)
            if depth > current_depth + 1:
                assets.extend(
                    generate_asset_tree(root_external_id + str(i), depth, children_per_node, current_depth + 1)
                )
    return assets


class TestAssetPoster:
    def test_validate_asset_hierarchy_asset_has_parent_id_and_parent_ref_id(self, cognite_client):
        assets = [Asset(external_id="1"), Asset(parent_external_id="1", parent_id=1, external_id="2")]
        with pytest.raises(AssertionError, match="has both"):
            _AssetPoster(assets, cognite_client.assets)

    def test_validate_asset_hierarchy_duplicate_ref_ids(self, cognite_client):
        assets = [Asset(external_id="1"), Asset(parent_external_id="1", external_id="1")]
        with pytest.raises(AssertionError, match="Duplicate"):
            _AssetPoster(assets, cognite_client.assets)

    def test_validate_asset_hierarchy__more_than_limit_only_resolved_assets(self, cognite_client):
        with set_request_limit(cognite_client.assets, 1):
            _AssetPoster(
                [Asset(external_id="a1", parent_id=1), Asset(external_id="a2", parent_id=2)], cognite_client.assets
            )

    def test_validate_asset_hierarchy_circular_dependencies(self, cognite_client):
        assets = [
            Asset(external_id="1", parent_external_id="3"),
            Asset(external_id="2", parent_external_id="1"),
            Asset(external_id="3", parent_external_id="2"),
        ]
        with set_request_limit(cognite_client.assets, 1):
            with pytest.raises(AssertionError, match="circular dependencies"):
                _AssetPoster(assets, cognite_client.assets)

    def test_validate_asset_hierarchy_self_dependency(self, cognite_client):
        assets = [Asset(external_id="1"), Asset(external_id="2", parent_external_id="2")]
        with set_request_limit(cognite_client.assets, 1):
            with pytest.raises(AssertionError, match="circular dependencies"):
                _AssetPoster(assets, cognite_client.assets)

    def test_initialize(self, cognite_client):
        assets = [
            Asset(external_id="1"),
            Asset(external_id="3", parent_external_id="1"),
            Asset(external_id="2", parent_external_id="1"),
            Asset(external_id="4", parent_external_id="2"),
        ]

        ap = _AssetPoster(assets, cognite_client.assets)
        assert OrderedDict({str(i): None for i in range(1, 5)}) == ap.remaining_external_ids
        assert {
            "1": {Asset(external_id="2", parent_external_id="1"), Asset(external_id="3", parent_external_id="1")},
            "2": {Asset(external_id="4", parent_external_id="2")},
            "3": set(),
            "4": set(),
        } == ap.external_id_to_children
        assert {"1": 3, "2": 1, "3": 0, "4": 0} == ap.external_id_to_descendent_count
        assert ap.assets_remaining() is True
        assert 0 == len(ap.posted_assets)
        assert ap.request_queue.empty()
        assert ap.response_queue.empty()
        assert {"1", "2", "3", "4"} == ap.remaining_external_ids_set

    def test_get_unblocked_assets__assets_unblocked_by_default_less_than_limit(self, cognite_client):
        assets = generate_asset_tree(root_external_id="0", depth=4, children_per_node=10)
        ap = _AssetPoster(assets=assets, client=cognite_client.assets)
        unblocked_assets_lists = ap._get_unblocked_assets()
        assert 1 == len(unblocked_assets_lists)
        assert 1000 == len(unblocked_assets_lists[0])

    def test_get_unblocked_assets__assets_unblocked_by_default_more_than_limit(self, cognite_client):
        assets = []
        for i in range(4):
            assets.extend(generate_asset_tree(root_external_id=str(i), depth=2, children_per_node=2))
        with set_request_limit(cognite_client.assets, 3):
            ap = _AssetPoster(assets=assets, client=cognite_client.assets)
            unblocked_assets_lists = ap._get_unblocked_assets()
        assert 4 == len(unblocked_assets_lists)
        for li in unblocked_assets_lists:
            assert 3 == len(li)

    def test_get_unblocked_assets_parent_ref_null_pointer(self, cognite_client):
        assets = [Asset(parent_external_id="1", external_id="2")]
        ap = _AssetPoster(assets, cognite_client.assets)
        asset_list = ap._get_unblocked_assets()
        assert len(asset_list) == 1

    @pytest.fixture
    def mock_post_asset_hierarchy(self, cognite_client, rsps):
        cognite_client.assets._config.max_workers = 1

        def request_callback(request):
            items = jsgz_load(request.body)["items"]
            response_assets = []
            for item in items:
                parent_id = None
                if "parentId" in item:
                    parent_id = item["parentId"]
                if "parentExternalId" in item:
                    parent_id = item["parentExternalId"] + "id"
                id = item.get("externalId", "root_") + "id"
                response_assets.append(
                    {
                        "id": id,
                        "parentId": parent_id,
                        "externalId": item["externalId"],
                        "parentExternalId": item.get("parentExternalId"),
                    }
                )
            return 200, {}, json.dumps({"items": response_assets})

        rsps.add_callback(
            rsps.POST,
            cognite_client.assets._get_base_url_with_base_path() + "/assets",
            callback=request_callback,
            content_type="application/json",
        )
        yield rsps
        cognite_client.assets._config.max_workers = 10

    @pytest.mark.parametrize(
        "limit, depth, children_per_node, expected_num_calls",
        [(100, 4, 10, 13), (9, 3, 9, 11), (100, 101, 1, 2), (1, 10, 1, 10)],
    )
    def test_post_hierarchy(
        self, cognite_client, limit, depth, children_per_node, expected_num_calls, mock_post_asset_hierarchy
    ):
        assets = generate_asset_tree(root_external_id="0", depth=depth, children_per_node=children_per_node)
        with set_request_limit(cognite_client.assets, limit):
            created_assets = cognite_client.assets.create_hierarchy(assets)

        assert len(assets) == len(created_assets)
        assert expected_num_calls - 1 <= len(mock_post_asset_hierarchy.calls) <= expected_num_calls + 1
        for asset in created_assets:
            if asset.id == "0id":
                assert asset.parent_id is None
            else:
                assert asset.id[:-3] == asset.parent_id[:-2]

    def test_post_assets_over_limit_only_resolved(self, cognite_client, mock_post_asset_hierarchy):
        with set_request_limit(cognite_client.assets, 1):
            _AssetPoster(
                [Asset(external_id="a1", parent_id=1), Asset(external_id="a2", parent_id=2)], cognite_client.assets
            ).post()
        assert 2 == len(mock_post_asset_hierarchy.calls)

    @pytest.fixture
    def mock_post_assets_failures(self, cognite_client, rsps):
        def request_callback(request):
            items = jsgz_load(request.body)["items"]
            response_assets = []
            item = items[0]
            parent_id = None
            if "parentId" in item:
                parent_id = item["parentId"]
            if "parentExternalId" in item:
                parent_id = item["parentExternalId"] + "id"
            id = item.get("refId", "root_") + "id"
            response_assets.append(
                {
                    "id": id,
                    "parentId": parent_id,
                    "externalId": item["externalId"],
                    "parentExternalId": item.get("parentExternalId"),
                }
            )

            if item["name"] == "400":
                return 400, {}, json.dumps({"error": {"message": "user error", "code": 400}})

            if item["name"] == "500":
                return 500, {}, json.dumps({"error": {"message": "internal server error", "code": 500}})

            return 200, {}, json.dumps({"items": response_assets})

        rsps.add_callback(
            rsps.POST,
            cognite_client.assets._get_base_url_with_base_path() + "/assets",
            callback=request_callback,
            content_type="application/json",
        )
        with set_request_limit(cognite_client.assets, 1):
            yield rsps

    def test_post_with_failures(self, cognite_client, mock_post_assets_failures):
        assets = [
            Asset(name="200", external_id="0"),
            Asset(name="200", parent_external_id="0", external_id="01"),
            Asset(name="400", parent_external_id="0", external_id="02"),
            Asset(name="200", parent_external_id="02", external_id="021"),
            Asset(name="200", parent_external_id="021", external_id="0211"),
            Asset(name="500", parent_external_id="0", external_id="03"),
            Asset(name="200", parent_external_id="03", external_id="031"),
        ]
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.assets.create_hierarchy(assets)

        assert {a.external_id for a in e.value.unknown} == {"03"}
        assert {a.external_id for a in e.value.failed} == {"02", "021", "0211", "031"}
        assert {a.external_id for a in e.value.successful} == {"0", "01"}


@pytest.fixture
def mock_assets_empty(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.assets._get_base_url_with_base_path()) + "/.+")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": []})
    yield rsps


@pytest.mark.dsl
class TestPandasIntegration:
    def test_asset_list_to_pandas(self, cognite_client, mock_assets_response):
        import pandas as pd

        df = cognite_client.assets.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_asset_list_to_pandas_empty(self, cognite_client, mock_assets_empty):
        import pandas as pd

        df = cognite_client.assets.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_asset_to_pandas(self, cognite_client, mock_assets_response):
        import pandas as pd

        asset = cognite_client.assets.retrieve(id=1)
        df = asset.to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert 1 == df.loc["id"][0]
        assert "metadata-value" == df.loc["metadata-key"][0]

    # need subtree here to get list, since to_pandas on a single Asset gives int for id, but on AssetList it gives int64
    def test_asset_id_from_to_pandas(self, cognite_client, mock_get_subtree):
        df = cognite_client.assets.retrieve_subtree(id=1).to_pandas()
        cognite_client.assets.retrieve(id=df.iloc[0]["id"])



import datetime
from unittest.mock import Mock, patch

import pytest
from oauthlib.oauth2 import InvalidClientIdError

from cognite.client import utils
from cognite.client.exceptions import CogniteAPIKeyError


def default_token_generator_args():
    return {
        "client_id": "azure-client-id",
        "client_secret": "azure-client-secret",
        "token_url": "https://login.microsoftonline.com/testingabc123/oauth2/v2.0/token",
        "token_scopes": ["https://greenfield.cognitedata.com/.default"],
        "token_custom_args": {},
    }


def setup_token_generator(generator_args=default_token_generator_args()):
    return utils._token_generator.TokenGenerator(
        generator_args["token_url"],
        generator_args["client_id"],
        generator_args["client_secret"],
        generator_args["token_scopes"],
        generator_args["token_custom_args"],
    )


class TestTokenGeneration:
    @patch("cognite.client.utils._token_generator.BackendApplicationClient")
    @patch("cognite.client.utils._token_generator.OAuth2Session")
    def test_all_token_environment_vars_set(self, mock_oauth_session, mock_backend_client):
        mock_backend_client().return_value = Mock()
        mock_oauth_session.fetch_token.return_value = {}
        generator = setup_token_generator()
        assert True == generator.token_params_set()

    @pytest.mark.parametrize("missing", ["token_url", "client_id", "client_secret", "token_scopes"])
    def test_missing_token_environment_vars(self, missing):
        generator_args = default_token_generator_args()
        generator_args[missing] = None
        generator = setup_token_generator(generator_args)
        assert False == generator.token_params_set()

    @patch("cognite.client.utils._token_generator.BackendApplicationClient")
    @patch("cognite.client.utils._token_generator.OAuth2Session")
    def test_access_token_generated(self, mock_oauth_session, mock_backend_client):
        mock_backend_client().return_value = Mock()
        mock_oauth_session().fetch_token.return_value = {
            "access_token": "azure_token",
            "expires_at": datetime.datetime.now().timestamp() + 1000,
        }
        generator = setup_token_generator()
        assert "azure_token" == generator.return_access_token()

    @patch("cognite.client.utils._token_generator.BackendApplicationClient")
    @patch("cognite.client.utils._token_generator.OAuth2Session")
    def test_access_token_not_generated(self, mock_oauth_session, mock_backend_client):
        mock_backend_client().return_value = Mock()
        mock_oauth_session().fetch_token.return_value = {}
        generator = setup_token_generator()
        with pytest.raises(
            CogniteAPIKeyError, match="Could not generate access token from provided token generation arguments"
        ):
            generator.return_access_token()

    @patch("cognite.client.utils._token_generator.BackendApplicationClient")
    @patch("cognite.client.utils._token_generator.OAuth2Session")
    def test_access_token_not_generated_due_to_error(self, mock_oauth_session, mock_backend_client):
        mock_backend_client().return_value = Mock()
        mock_oauth_session().fetch_token.side_effect = InvalidClientIdError()
        with pytest.raises(
            CogniteAPIKeyError,
            match="Error generating access token: invalid_request, 400, Invalid client_id parameter value.",
        ):
            args = default_token_generator_args()
            utils._token_generator.TokenGenerator(
                args["token_url"],
                args["client_id"],
                args["client_secret"],
                args["token_scopes"],
                args["token_custom_args"],
            )

    def test_access_token_not_generated_missing_args(self):
        generator_args = default_token_generator_args()
        generator_args["client_secret"] = None
        generator = setup_token_generator(generator_args)
        with pytest.raises(
            CogniteAPIKeyError, match="Could not generate access token - missing token generation arguments"
        ):
            generator.return_access_token()

    @patch("cognite.client.utils._token_generator.BackendApplicationClient")
    @patch("cognite.client.utils._token_generator.OAuth2Session")
    def test_access_token_expired(self, mock_oauth_session, mock_backend_client):
        mock_backend_client().return_value = Mock()
        mock_oauth_session().fetch_token.side_effect = [
            {"access_token": "azure_token_expired", "expires_at": datetime.datetime.now().timestamp() - 1000},
            {"access_token": "azure_token_refreshed", "expires_at": datetime.datetime.now().timestamp() + 1000},
        ]
        generator = setup_token_generator()
        assert "azure_token_expired" == generator._access_token
        assert "azure_token_refreshed" == generator.return_access_token()



import re

import pytest

from cognite.client.data_classes import Asset, EntityMatchingModel, TimeSeries
from cognite.client.exceptions import ModelFailedException
from tests.utils import jsgz_load


@pytest.fixture
def mock_fit(rsps, cognite_client):
    response_body = {"id": 123, "status": "Queued", "createdTime": 42}
    rsps.add(
        rsps.POST,
        cognite_client.entity_matching._get_base_url_with_base_path()
        + cognite_client.entity_matching._RESOURCE_PATH
        + "/",
        status=200,
        json=response_body,
    )
    yield rsps


@pytest.fixture
def mock_status_ok(rsps, cognite_client):
    response_body = {"id": 123, "status": "Completed", "createdTime": 42, "statusTime": 456, "startTime": 789}
    rsps.add(
        rsps.GET,
        re.compile(
            f"{cognite_client.entity_matching._get_base_url_with_base_path()}{cognite_client.entity_matching._RESOURCE_PATH}/\\d+"
        ),
        status=200,
        json=response_body,
    )
    yield rsps


@pytest.fixture
def mock_retrieve(rsps, cognite_client):
    response_body = {
        "items": [{"id": 123, "status": "Completed", "createdTime": 42, "statusTime": 456, "startTime": 789}]
    }
    rsps.add(
        rsps.POST,
        re.compile(
            f"{cognite_client.entity_matching._get_base_url_with_base_path()}{cognite_client.entity_matching._RESOURCE_PATH}/byids"
        ),
        status=200,
        json=response_body,
    )
    yield rsps


@pytest.fixture
def mock_status_failed(rsps, cognite_client):
    response_body = {"id": 123, "status": "Failed", "errorMessage": "error message"}
    rsps.add(
        rsps.GET,
        re.compile(
            f"{cognite_client.entity_matching._get_base_url_with_base_path()}{cognite_client.entity_matching._RESOURCE_PATH}/\\d+"
        ),
        status=200,
        json=response_body,
    )
    yield rsps


@pytest.fixture
def mock_status_rules_ok(rsps, cognite_client):
    response_body = {"jobId": 456, "status": "Completed", "items": [1]}
    rsps.add(
        rsps.GET,
        re.compile(
            f"{cognite_client.entity_matching._get_base_url_with_base_path()}{cognite_client.entity_matching._RESOURCE_PATH}/rules/\\d+"
        ),
        status=200,
        json=response_body,
    )
    yield rsps


class TestEntityMatching:
    def test_fit(self, cognite_client, mock_fit, mock_status_ok):
        entities_from = [{"id": 1, "name": "xx"}]
        entities_to = [{"id": 2, "name": "yy"}]
        model = cognite_client.entity_matching.fit(
            sources=entities_from, targets=entities_to, true_matches=[(1, 2)], feature_type="bigram"
        )
        assert isinstance(model, EntityMatchingModel)
        assert "EntityMatchingModel(id: 123,status: Queued,error: None)" == str(model)
        assert 42 == model.created_time
        model.wait_for_completion()
        assert "Completed" == model.status
        assert 123 == model.id
        assert 42 == model.created_time
        assert 456 == model.status_time
        assert 789 == model.start_time

        n_fit_calls = 0
        n_status_calls = 0
        for call in mock_fit.calls:
            if call.request.method == "POST":
                n_fit_calls += 1
                assert {
                    "sources": entities_from,
                    "targets": entities_to,
                    "trueMatches": [{"sourceId": 1, "targetId": 2}],
                    "featureType": "bigram",
                    "ignoreMissingFields": False,
                    "matchFields": None,
                    "name": None,
                    "description": None,
                    "externalId": None,
                    "classifier": None,
                } == jsgz_load(call.request.body)
            else:
                n_status_calls += 1
                assert "/123" in call.request.url
        assert 1 == n_fit_calls
        assert 1 == n_status_calls

    def test_ml_fit(self, cognite_client, mock_fit, mock_status_ok):
        # fit_ml should produce the same output as fit. Will eventually be removed
        entities_from = [{"id": 1, "name": "xx"}]
        entities_to = [{"id": 2, "name": "yy"}]
        model = cognite_client.entity_matching.fit(
            sources=entities_from, targets=entities_to, true_matches=[(1, 2)], feature_type="bigram"
        )
        assert isinstance(model, EntityMatchingModel)
        assert "EntityMatchingModel(id: 123,status: Queued,error: None)" == str(model)
        assert 42 == model.created_time
        model.wait_for_completion()
        assert "Completed" == model.status
        assert 123 == model.id

    def test_fit_cognite_resource(self, cognite_client, mock_fit):
        entities_from = [TimeSeries(id=1, name="x", metadata={"ka": "va"})]
        entities_to = [Asset(id=1, external_id="abc", name="x")]
        cognite_client.entity_matching.fit(
            sources=entities_from, targets=entities_to, true_matches=[(1, "abc")], feature_type="bigram"
        )
        assert {
            "sources": [{"id": 1, "name": "x", "metadata.ka": "va"}],
            "targets": [entities_to[0].dump(camel_case=True)],
            "trueMatches": [{"sourceId": 1, "targetExternalId": "abc"}],
            "featureType": "bigram",
            "ignoreMissingFields": False,
            "matchFields": None,
            "name": None,
            "description": None,
            "externalId": None,
            "classifier": None,
        } == jsgz_load(mock_fit.calls[0].request.body)

    def test_fit_fails(self, cognite_client, mock_fit, mock_status_failed):
        entities_from = [{"id": 1, "name": "xx"}]
        entities_to = [{"id": 2, "name": "yy"}]
        model = cognite_client.entity_matching.fit(sources=entities_from, targets=entities_to)
        with pytest.raises(ModelFailedException) as exc_info:
            model.wait_for_completion()
        assert exc_info.type is ModelFailedException
        assert 123 == exc_info.value.id
        assert "error message" == exc_info.value.error_message
        assert "EntityMatchingModel 123 failed with error 'error message'" == str(exc_info.value)

    def test_retrieve(self, cognite_client, mock_retrieve):
        model = cognite_client.entity_matching.retrieve(id=123)
        assert isinstance(model, EntityMatchingModel)
        assert "Completed" == model.status
        assert 123 == model.id



import random
import time
from unittest import mock

import pytest

from cognite.client import utils
from cognite.client.data_classes import Asset, AssetFilter, AssetList, AssetUpdate
from cognite.client.exceptions import CogniteNotFoundError
from tests.utils import set_request_limit


@pytest.fixture
def new_asset(cognite_client):
    ts = cognite_client.assets.create(Asset(name="any", description="haha", metadata={"a": "b"}))
    yield ts
    cognite_client.assets.delete(id=ts.id)
    assert cognite_client.assets.retrieve(ts.id) is None


def generate_asset_tree(root_external_id: str, depth: int, children_per_node: int, current_depth=1):
    assert 1 <= children_per_node <= 10, "children_per_node must be between 1 and 10"
    assets = []
    if current_depth == 1:
        assets = [Asset(external_id=root_external_id, name=root_external_id)]
    if depth > current_depth:
        for i in range(children_per_node):
            external_id = "{}{}".format(root_external_id, i)
            asset = Asset(parent_external_id=root_external_id, external_id=external_id, name=external_id)
            assets.append(asset)
            if depth > current_depth + 1:
                assets.extend(
                    generate_asset_tree(root_external_id + str(i), depth, children_per_node, current_depth + 1)
                )
    return assets


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.assets, "_post", wraps=cognite_client.assets._post) as _:
        yield


@pytest.fixture
def new_asset_hierarchy(cognite_client, post_spy):
    random_prefix = "test_{}_".format(utils._auxiliary.random_string(10))
    assets = generate_asset_tree(random_prefix + "0", depth=5, children_per_node=5)

    with set_request_limit(cognite_client.assets, 50):
        cognite_client.assets.create_hierarchy(assets)

    assert 20 < cognite_client.assets._post.call_count < 30

    ext_ids = [a.external_id for a in assets]
    yield random_prefix, ext_ids

    cognite_client.assets.delete(external_id=random_prefix + "0", recursive=True)


@pytest.fixture
def root_test_asset(cognite_client):
    for _ in range(5):  # remove after race condition CDF-10303 is fixed
        for asset in cognite_client.assets(root=True):
            if asset.name.startswith("test__"):
                return asset
        time.sleep(1)
    assert False  # should never be reached


@pytest.fixture
def new_root_asset(cognite_client):
    external_id = "my_root_{}".format(utils._auxiliary.random_string(10))
    root = Asset(external_id=external_id, name="my_root")
    root = cognite_client.assets.create(root)
    yield root
    cognite_client.assets.delete(external_id=external_id, recursive=True)
    assert cognite_client.assets.retrieve(external_id=external_id) is None


class TestAssetsAPI:
    def test_get(self, cognite_client):
        res = cognite_client.assets.list(limit=1)
        assert res[0] == cognite_client.assets.retrieve(res[0].id)

    def test_retrieve_unknown(self, cognite_client):
        res = cognite_client.assets.list(limit=1)
        with pytest.raises(CogniteNotFoundError):
            cognite_client.assets.retrieve_multiple(ids=[res[0].id], external_ids=["this does not exist"])
        retr = cognite_client.assets.retrieve_multiple(
            ids=[res[0].id], external_ids=["this does not exist"], ignore_unknown_ids=True
        )
        assert 1 == len(retr)

    def test_list(self, cognite_client, post_spy):
        with set_request_limit(cognite_client.assets, 10):
            res = cognite_client.assets.list(limit=20)

        assert 20 == len(res)
        assert 2 == cognite_client.assets._post.call_count

    def test_partitioned_list(self, cognite_client, post_spy):
        # stop race conditions by cutting off max created time
        maxtime = int(time.time() - 3600) * 1000
        res_flat = cognite_client.assets.list(limit=None, created_time={"max": maxtime})
        res_part = cognite_client.assets.list(partitions=8, limit=None, created_time={"max": maxtime})
        assert len(res_flat) > 0
        assert len(res_flat) == len(res_part)
        assert {a.id for a in res_flat} == {a.id for a in res_part}

    def test_compare_partitioned_gen_and_list(self, cognite_client, post_spy):
        # stop race conditions by cutting off max created time
        maxtime = int(time.time() - 3600) * 1000
        res_generator = cognite_client.assets(partitions=8, limit=None, created_time={"max": maxtime})
        res_list = cognite_client.assets.list(partitions=8, limit=None, created_time={"max": maxtime})
        assert {a.id for a in res_generator} == {a.id for a in res_list}

    def test_list_with_aggregated_properties_param(self, cognite_client, post_spy):
        res = cognite_client.assets.list(limit=10, aggregated_properties=["child_count"])
        for asset in res:
            assert {"childCount"} == asset.aggregates.keys()
            assert isinstance(asset.aggregates["childCount"], int)

    def test_aggregate(self, cognite_client, new_asset):
        res = cognite_client.assets.aggregate(filter=AssetFilter(name="test__asset_0"))
        assert res[0].count > 0

    def test_search(self, cognite_client):
        res = cognite_client.assets.search(name="test__asset_0", filter=AssetFilter(name="test__asset_0"))
        assert len(res) > 0

    def test_search_query(self, cognite_client):
        res = cognite_client.assets.search(query="test asset 0", limit=5)
        assert len(res) > 0

    def test_update(self, cognite_client, new_asset):
        assert new_asset.metadata == {"a": "b"}
        assert new_asset.description == "haha"
        update_asset = AssetUpdate(new_asset.id).name.set("newname").metadata.set(None).description.set(None)
        res = cognite_client.assets.update(update_asset)
        assert "newname" == res.name
        assert res.metadata == {}
        assert res.description is None

    def test_delete_with_nonexisting(self, cognite_client):
        a = cognite_client.assets.create(Asset(name="any"))
        cognite_client.assets.delete(id=a.id, external_id="this asset does not exist", ignore_unknown_ids=True)
        assert cognite_client.assets.retrieve(id=a.id) is None

    def test_post_asset_hierarchy(self, cognite_client, new_asset_hierarchy):
        prefix, ext_ids = new_asset_hierarchy
        posted_assets = cognite_client.assets.retrieve_multiple(external_ids=ext_ids)
        external_id_to_id = {a.external_id: a.id for a in posted_assets}

        for asset in posted_assets:
            if asset.external_id == prefix + "0":
                assert asset.parent_id is None
            else:
                assert asset.parent_id == external_id_to_id[asset.external_id[:-1]]

    def test_get_subtree(self, cognite_client, root_test_asset):
        assert isinstance(cognite_client.assets.retrieve_subtree(id=random.randint(1, 10)), AssetList)
        assert 0 == len(cognite_client.assets.retrieve_subtree(external_id="non_existing_asset"))
        assert 0 == len(cognite_client.assets.retrieve_subtree(id=random.randint(1, 10)))
        assert 781 == len(cognite_client.assets.retrieve_subtree(root_test_asset.id))
        subtree = cognite_client.assets.retrieve_subtree(root_test_asset.id, depth=1)
        assert 6 == len(subtree)
        assert all(subtree.get(id=a.id) is not None for a in subtree)  # test if .get works on result

    def test_create_asset_hierarchy_parent_external_id_not_in_request(self, cognite_client, new_root_asset):
        root = new_root_asset
        children = generate_asset_tree(
            root_external_id=root.external_id, depth=5, children_per_node=10, current_depth=2
        )

        cognite_client.assets.create_hierarchy(children)

        external_ids = [asset.external_id for asset in children] + [root.external_id]
        posted_assets = cognite_client.assets.retrieve_multiple(external_ids=external_ids)
        external_id_to_id = {a.external_id: a.id for a in posted_assets}
        for asset in posted_assets:
            if asset.external_id == root.external_id:
                assert asset.parent_id is None
            else:
                assert asset.parent_id == external_id_to_id[asset.external_id[:-1]]



from unittest import mock

import numpy as np
import pytest

from cognite.client.data_classes import Sequence, SequenceData, SequenceDataList


@pytest.fixture(scope="session")
def named_long_str(cognite_client):
    seq = cognite_client.sequences.retrieve(external_id="named_long_str")
    assert isinstance(seq, Sequence)
    yield seq


@pytest.fixture(scope="session")
def string200(cognite_client):
    seq = cognite_client.sequences.retrieve(external_id="string200")
    assert isinstance(seq, Sequence)
    yield seq


@pytest.fixture(scope="session")
def small_sequence(cognite_client):
    seq = cognite_client.sequences.retrieve(external_id="small")
    assert isinstance(seq, Sequence)
    yield seq


@pytest.fixture(scope="session")
def pretend_timeseries(cognite_client):
    seq = cognite_client.sequences.retrieve(external_id="pretend_timeseries")
    assert isinstance(seq, Sequence)
    yield seq


@pytest.fixture(scope="session")
def new_seq(cognite_client):
    seq = cognite_client.sequences.create(Sequence(columns=[{"valueType": "STRING"}]))
    yield seq
    cognite_client.sequences.delete(id=seq.id)
    assert cognite_client.sequences.retrieve(seq.id) is None


@pytest.fixture(scope="session")
def new_small_seq(cognite_client, small_sequence):
    seq = cognite_client.sequences.create(Sequence(columns=small_sequence.columns))
    yield seq
    cognite_client.sequences.delete(id=seq.id)
    assert cognite_client.sequences.retrieve(seq.id) is None


@pytest.fixture(scope="session")
def new_seq_long(cognite_client):
    seq = cognite_client.sequences.create(Sequence(columns=[{"valueType": "LONG", "externalId": "a"}]))
    yield seq
    cognite_client.sequences.delete(id=seq.id)
    assert cognite_client.sequences.retrieve(seq.id) is None


@pytest.fixture(scope="session")
def new_seq_mixed(cognite_client):
    seq = cognite_client.sequences.create(Sequence(columns=[{"valueType": "DOUBLE"}, {"valueType": "STRING"}]))
    yield seq
    cognite_client.sequences.delete(id=seq.id)
    assert cognite_client.sequences.retrieve(seq.id) is None


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.sequences.data, "_post", wraps=cognite_client.sequences.data._post) as _:
        yield


class TestSequencesDataAPI:
    def test_retrieve(self, cognite_client, small_sequence):
        dps = cognite_client.sequences.data.retrieve(id=small_sequence.id, start=0, end=None)

        assert isinstance(dps, SequenceData)
        assert len(dps) > 0

    def test_retrieve_multi(self, cognite_client, small_sequence, pretend_timeseries):
        dps = cognite_client.sequences.data.retrieve(
            id=[small_sequence.id], external_id=pretend_timeseries.external_id, start=0, end=None
        )
        assert isinstance(dps, SequenceDataList)
        assert len(dps[0]) > 0
        assert len(dps[1]) > 0
        assert small_sequence.id == dps[0].id
        assert pretend_timeseries.external_id == dps[1].external_id

    def test_retrieve_multi_dataframe(self, cognite_client, small_sequence, pretend_timeseries):
        df = cognite_client.sequences.data.retrieve_dataframe(
            id=[small_sequence.id, pretend_timeseries.id], start=0, end=None, column_names="id"
        )
        assert df.shape[0] > 0
        assert 3 == df.shape[1]
        assert all([str(small_sequence.id), str(small_sequence.id), str(pretend_timeseries.id)] == df.columns)

    def test_retrieve_dataframe(self, cognite_client, small_sequence):
        df = cognite_client.sequences.data.retrieve_dataframe(id=small_sequence.id, start=0, end=5)
        assert df.shape[0] == 4
        assert df.shape[1] == 2
        assert np.diff(df.index).all()

    def test_insert_dataframe(self, cognite_client, small_sequence, new_small_seq):
        df = cognite_client.sequences.data.retrieve_dataframe(id=small_sequence.id, start=0, end=5)
        cognite_client.sequences.data.insert_dataframe(df, id=new_small_seq.id)

    def test_insert(self, cognite_client, new_seq):
        data = {i: ["str"] for i in range(1, 61)}
        cognite_client.sequences.data.insert(rows=data, column_external_ids=new_seq.column_external_ids, id=new_seq.id)

    def test_insert_raw(self, cognite_client, new_seq_long):
        data = [{"rowNumber": i, "values": [2 * i]} for i in range(1, 61)]
        cognite_client.sequences.data.insert(
            rows=data, column_external_ids=new_seq_long.column_external_ids, id=new_seq_long.id
        )

    def test_insert_implicit_rows(self, cognite_client, new_seq_mixed):
        data = {i: [i, "str"] for i in range(1, 10)}
        cognite_client.sequences.data.insert(data, id=new_seq_mixed.id, column_external_ids=["column0", "column1"])

    def test_insert_copy(self, cognite_client, small_sequence, new_small_seq):
        data = cognite_client.sequences.data.retrieve(id=small_sequence.id, start=0, end=5)
        cognite_client.sequences.data.insert(rows=data, id=new_small_seq.id, column_external_ids=None)

    def test_delete_multiple(self, cognite_client, new_seq):
        cognite_client.sequences.data.delete(rows=[1, 2, 42, 3524], id=new_seq.id)

    def test_retrieve_paginate(self, cognite_client, string200, post_spy):
        data = cognite_client.sequences.data.retrieve(start=1, end=996, id=string200.id)
        assert 200 == len(data.values[0])
        assert 995 == len(data)
        assert 4 == cognite_client.sequences.data._post.call_count  # around 300 rows per request for this case

    def test_retrieve_paginate_max(self, cognite_client, pretend_timeseries, post_spy):
        data = cognite_client.sequences.data.retrieve(start=0, end=None, id=pretend_timeseries.id)
        assert 1 == len(data.values[0])
        assert 54321 == len(data)
        assert 6 == cognite_client.sequences.data._post.call_count  # 10k rows each of 54321 rows

    def test_retrieve_paginate_limit_small(self, cognite_client, pretend_timeseries, post_spy):
        data = cognite_client.sequences.data.retrieve(start=0, end=None, id=pretend_timeseries.id, limit=23)
        assert 1 == len(data.values[0])
        assert 23 == len(data)
        assert 1 == cognite_client.sequences.data._post.call_count  # 10k rows each of 54321 rows

    def test_retrieve_paginate_limit_paged(self, cognite_client, pretend_timeseries, post_spy):
        data = cognite_client.sequences.data.retrieve_dataframe(
            start=0, end=None, id=pretend_timeseries.id, limit=40023
        )
        assert 1 == data.shape[1]
        assert 40023 == data.shape[0]
        assert 5 == cognite_client.sequences.data._post.call_count

    def test_retrieve_one_column(self, cognite_client, named_long_str):
        dps = cognite_client.sequences.data.retrieve(
            id=named_long_str.id, start=42, end=43, column_external_ids=["strcol"]
        )
        assert 1 == len(dps)
        assert 1 == len(dps.column_external_ids)
        assert isinstance(dps.values[0][0], str)

    def test_retrieve_mixed(self, cognite_client, named_long_str):
        dps = cognite_client.sequences.data.retrieve(id=named_long_str.id, start=42, end=43)
        assert 1 == len(dps)
        assert isinstance(dps.values[0][0], int)
        assert isinstance(dps.values[0][1], str)
        assert 1 == len(dps.get_column("longcol"))
        assert isinstance(dps.get_column("longcol")[0], int)
        assert isinstance(dps.get_column(external_id="strcol")[0], str)
        with pytest.raises(ValueError):
            dps.get_column("missingcol")

    def test_retrieve_paginate_end_coinciding_with_page(self, cognite_client, string200, post_spy):
        cognite_client.sequences.data.retrieve(start=1, end=118, id=string200.id)
        assert 1 == cognite_client.sequences.data._post.call_count

    def test_delete_range(self, cognite_client, new_seq_long):
        data = [(i, [10 * i]) for i in [1, 2, 3, 5, 8, 13, 21, 34]]
        cognite_client.sequences.data.insert(
            column_external_ids=new_seq_long.column_external_ids, rows=data, id=new_seq_long.id
        )
        cognite_client.sequences.data.delete_range(start=4, end=15, id=new_seq_long.id)
        dps = cognite_client.sequences.data.retrieve(start=0, end=None, id=new_seq_long.id)
        # potential delay, so can't assert, but tested in notebook
        # assert [10, 20, 30, 210, 340] == [d[0] for d in dps.values]
        # assert [1, 2, 3, 21, 34] == dps.row_numbers



import os

import pytest
from msal import PublicClientApplication

from cognite.client import CogniteClient

REDIRECT_PORT = 53000


def make_cognite_client_with_interactive_flow() -> CogniteClient:
    authority_url = os.environ["COGNITE_AUTHORITY_URL"]
    client_id = os.environ["COGNITE_CLIENT_ID"]
    scopes = os.environ.get("COGNITE_TOKEN_SCOPES", "").split(",")
    app = PublicClientApplication(client_id=client_id, authority=authority_url)
    creds = app.acquire_token_interactive(scopes=scopes, port=REDIRECT_PORT)
    return CogniteClient(token=creds["access_token"])


@pytest.fixture(scope="session")
def cognite_client() -> CogniteClient:
    login_flow = os.environ["LOGIN_FLOW"].lower()
    if login_flow == "client_credentials":
        return CogniteClient()
    elif login_flow == "interactive":
        return make_cognite_client_with_interactive_flow()
    else:
        raise ValueError("Environment variable LOGIN_FLOW must be set to either 'client_credentials' or 'interactive'")



import re

import pytest

from cognite.client._api.three_d import (
    ThreeDAssetMapping,
    ThreeDAssetMappingList,
    ThreeDModel,
    ThreeDModelList,
    ThreeDModelRevision,
    ThreeDModelRevisionList,
    ThreeDModelRevisionUpdate,
    ThreeDModelUpdate,
    ThreeDNodeList,
)
from cognite.client.exceptions import CogniteAPIError
from tests.utils import jsgz_load


@pytest.fixture
def mock_3d_model_response(rsps, cognite_client):
    response_body = {"items": [{"name": "My Model", "id": 1000, "createdTime": 0}]}
    url_pattern = re.compile(re.escape(cognite_client.three_d._get_base_url_with_base_path()) + "/3d/models.*")
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    rsps.assert_all_requests_are_fired = False
    yield rsps


@pytest.fixture
def mock_retrieve_3d_model_response(rsps, cognite_client):
    response_body = {"name": "My Model", "id": 1000, "createdTime": 0}
    rsps.add(
        rsps.GET, cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1", status=200, json=response_body
    )
    yield rsps


class Test3DModels:
    def test_list(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.list(published=True, limit=100)
        assert isinstance(res, ThreeDModelList)
        assert mock_3d_model_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_published_default(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.list(limit=100)
        assert isinstance(res, ThreeDModelList)
        assert mock_3d_model_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert "published" not in mock_3d_model_response.calls[0].request.path_url

    def test_update_with_update_object(self, cognite_client, mock_3d_model_response):
        update = ThreeDModelUpdate(id=1).name.set("bla")
        res = cognite_client.three_d.models.update(update)
        assert {"id": 1, "update": {"name": {"set": "bla"}}} == jsgz_load(mock_3d_model_response.calls[0].request.body)[
            "items"
        ][0]
        assert mock_3d_model_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_resource_object(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.update(ThreeDModel(id=1, name="bla", created_time=123))
        assert {"id": 1, "update": {"name": {"set": "bla"}}} == jsgz_load(mock_3d_model_response.calls[0].request.body)[
            "items"
        ][0]
        assert mock_3d_model_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.delete(id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_3d_model_response.calls[0].request.body)
        assert res is None
        res = cognite_client.three_d.models.delete(id=[1])
        assert {"items": [{"id": 1}]} == jsgz_load(mock_3d_model_response.calls[1].request.body)
        assert res is None

    def test_retrieve(self, cognite_client, mock_retrieve_3d_model_response):
        res = cognite_client.three_d.models.retrieve(id=1)
        assert isinstance(res, ThreeDModel)
        assert mock_retrieve_3d_model_response.calls[0].response.json() == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.create(name="My Model")
        assert isinstance(res, ThreeDModel)
        assert jsgz_load(mock_3d_model_response.calls[0].request.body) == {"items": [{"name": "My Model"}]}
        assert mock_3d_model_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_3d_model_response):
        res = cognite_client.three_d.models.create(name=["My Model"])
        assert isinstance(res, ThreeDModelList)
        assert jsgz_load(mock_3d_model_response.calls[0].request.body) == {"items": [{"name": "My Model"}]}
        assert mock_3d_model_response.calls[0].response.json()["items"] == res.dump(camel_case=True)


@pytest.fixture
def mock_3d_model_revision_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "id": 1,
                "fileId": 1000,
                "published": False,
                "rotation": [0, 0, 0],
                "camera": {"target": [0, 0, 0], "position": [0, 0, 0]},
                "status": "Done",
                "thumbnailThreedFileId": 1000,
                "thumbnailUrl": "https://api.cognitedata.com/api/v1/project/myproject/3d/files/1000",
                "assetMappingCount": 0,
                "createdTime": 0,
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.three_d._get_base_url_with_base_path()) + "/3d/models/1/revisions.*"
    )
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    rsps.assert_all_requests_are_fired = False
    yield rsps


@pytest.fixture
def mock_retrieve_3d_model_revision_response(rsps, cognite_client):
    res = {
        "id": 1000,
        "fileId": 1000,
        "published": False,
        "rotation": [0, 0, 0],
        "camera": {"target": [0, 0, 0], "position": [0, 0, 0]},
        "status": "Done",
        "thumbnailThreedFileId": 1000,
        "thumbnailUrl": "https://api.cognitedata.com/api/v1/project/myproject/3d/files/1000",
        "assetMappingCount": 0,
        "createdTime": 0,
    }
    rsps.add(
        rsps.GET,
        cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1",
        status=200,
        json=res,
    )
    yield rsps


@pytest.fixture
def mock_3d_model_revision_thumbnail_response(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1/thumbnail",
        status=200,
        json={},
    )
    yield rsps


@pytest.fixture
def mock_3d_model_revision_node_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "id": 1,
                "treeIndex": 3,
                "parentId": 2,
                "depth": 2,
                "name": "Node name",
                "subtreeSize": 4,
                "boundingBox": {"max": [0, 0, 0], "min": [0, 0, 0]},
            }
        ]
    }
    rsps.add(
        rsps.GET,
        cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1/nodes",
        status=200,
        json=response_body,
    )
    rsps.add(
        rsps.POST,
        cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1/nodes/list",
        status=200,
        json=response_body,
    )
    rsps.add(
        rsps.GET,
        cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1/nodes/ancestors",
        status=200,
        json=response_body,
    )
    rsps.assert_all_requests_are_fired = False
    yield rsps


class Test3DModelRevisions:
    def test_list(self, cognite_client, mock_3d_model_revision_response):
        res = cognite_client.three_d.revisions.list(model_id=1, published=True, limit=100)
        assert isinstance(res, ThreeDModelRevisionList)
        assert mock_3d_model_revision_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_with_update_object(self, cognite_client, mock_3d_model_revision_response):
        update = ThreeDModelRevisionUpdate(id=1).published.set(False)
        cognite_client.three_d.revisions.update(1, update)
        assert {"id": 1, "update": {"published": {"set": False}}} == jsgz_load(
            mock_3d_model_revision_response.calls[0].request.body
        )["items"][0]

    def test_update_with_resource_object(self, cognite_client, mock_3d_model_revision_response):
        cognite_client.three_d.revisions.update(1, ThreeDModelRevision(id=1, published=False, created_time=123))
        assert {"id": 1, "update": {"published": {"set": False}}} == jsgz_load(
            mock_3d_model_revision_response.calls[0].request.body
        )["items"][0]

    def test_delete(self, cognite_client, mock_3d_model_revision_response):
        res = cognite_client.three_d.revisions.delete(1, id=1)
        assert {"items": [{"id": 1}]} == jsgz_load(mock_3d_model_revision_response.calls[0].request.body)
        assert res is None
        res = cognite_client.three_d.revisions.delete(1, id=[1])
        assert {"items": [{"id": 1}]} == jsgz_load(mock_3d_model_revision_response.calls[1].request.body)
        assert res is None

    def test_retrieve(self, cognite_client, mock_retrieve_3d_model_revision_response):
        res = cognite_client.three_d.revisions.retrieve(model_id=1, id=1)
        assert isinstance(res, ThreeDModelRevision)
        assert mock_retrieve_3d_model_revision_response.calls[0].response.json() == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_3d_model_revision_response):
        res = cognite_client.three_d.revisions.create(model_id=1, revision=ThreeDModelRevision(file_id=123))
        assert isinstance(res, ThreeDModelRevision)
        assert {"items": [{"fileId": 123}]} == jsgz_load(mock_3d_model_revision_response.calls[0].request.body)
        assert mock_3d_model_revision_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_3d_model_revision_response):
        res = cognite_client.three_d.revisions.create(model_id=1, revision=[ThreeDModelRevision(file_id=123)])
        assert isinstance(res, ThreeDModelRevisionList)
        assert {"items": [{"fileId": 123}]} == jsgz_load(mock_3d_model_revision_response.calls[0].request.body)
        assert mock_3d_model_revision_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_update_thumbnail(self, cognite_client, mock_3d_model_revision_thumbnail_response):
        res = cognite_client.three_d.revisions.update_thumbnail(model_id=1, revision_id=1, file_id=1)
        assert {"fileId": 1} == jsgz_load(mock_3d_model_revision_thumbnail_response.calls[0].request.body)
        assert res is None

    def test_list_3d_nodes(self, cognite_client, mock_3d_model_revision_node_response):
        res = cognite_client.three_d.revisions.list_nodes(model_id=1, revision_id=1, node_id=None, depth=None, limit=10)
        assert isinstance(res, ThreeDNodeList)
        assert mock_3d_model_revision_node_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_filter_3d_nodes(self, cognite_client, mock_3d_model_revision_node_response):
        res = cognite_client.three_d.revisions.filter_nodes(
            model_id=1, revision_id=1, properties={"Item": {"Type": ["Group"]}}, limit=10
        )
        assert isinstance(res, ThreeDNodeList)
        assert mock_3d_model_revision_node_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_3d_ancestor_nodes(self, cognite_client, mock_3d_model_revision_node_response):
        res = cognite_client.three_d.revisions.list_ancestor_nodes(model_id=1, revision_id=1, node_id=None, limit=10)
        assert isinstance(res, ThreeDNodeList)
        assert mock_3d_model_revision_node_response.calls[0].response.json()["items"] == res.dump(camel_case=True)


class Test3DFiles:
    @pytest.fixture
    def mock_3d_files_response(self, cognite_client, rsps):
        rsps.add(rsps.GET, cognite_client.three_d._get_base_url_with_base_path() + "/3d/files/1", body="bla")

    def test_retrieve(self, cognite_client, mock_3d_files_response):
        assert b"bla" == cognite_client.three_d.files.retrieve(1)


class Test3DAssetMappings:
    @pytest.fixture
    def mock_3d_asset_mappings_response(self, cognite_client, rsps):
        response_body = {"items": [{"nodeId": 1003, "assetId": 3001, "treeIndex": 5, "subtreeSize": 7}]}
        url_pattern = re.compile(
            re.escape(cognite_client.three_d._get_base_url_with_base_path()) + "/3d/models/1/revisions/1/mappings.*"
        )

        rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
        rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
        rsps.assert_all_requests_are_fired = False
        yield rsps

    def test_list(self, cognite_client, mock_3d_asset_mappings_response):
        res = cognite_client.three_d.asset_mappings.list(
            model_id=1, revision_id=1, node_id=None, asset_id=None, limit=None
        )
        assert isinstance(res, ThreeDAssetMappingList)
        assert mock_3d_asset_mappings_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_3d_asset_mappings_response):
        res = cognite_client.three_d.asset_mappings.create(
            model_id=1, revision_id=1, asset_mapping=ThreeDAssetMapping(node_id=1, asset_id=1)
        )
        assert isinstance(res, ThreeDAssetMapping)
        assert mock_3d_asset_mappings_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_3d_asset_mappings_response):
        res = cognite_client.three_d.asset_mappings.create(
            model_id=1, revision_id=1, asset_mapping=[ThreeDAssetMapping(node_id=1, asset_id=1)]
        )
        assert isinstance(res, ThreeDAssetMappingList)
        assert mock_3d_asset_mappings_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_3d_asset_mappings_response):
        res = cognite_client.three_d.asset_mappings.delete(
            model_id=1, revision_id=1, asset_mapping=ThreeDAssetMapping(1, 1)
        )
        assert res is None
        assert [{"nodeId": 1, "assetId": 1}] == jsgz_load(mock_3d_asset_mappings_response.calls[0].request.body)[
            "items"
        ]

    def test_delete_multiple(self, cognite_client, mock_3d_asset_mappings_response):
        res = cognite_client.three_d.asset_mappings.delete(
            model_id=1, revision_id=1, asset_mapping=[ThreeDAssetMapping(1, 1)]
        )
        assert res is None
        assert [{"nodeId": 1, "assetId": 1}] == jsgz_load(mock_3d_asset_mappings_response.calls[0].request.body)[
            "items"
        ]

    def test_delete_fails(self, cognite_client, rsps):
        rsps.add(
            rsps.POST,
            cognite_client.three_d._get_base_url_with_base_path() + "/3d/models/1/revisions/1/mappings/delete",
            status=500,
            json={"error": {"message": "Server Error", "code": 500}},
        )
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.three_d.asset_mappings.delete(
                model_id=1, revision_id=1, asset_mapping=[ThreeDAssetMapping(1, 1)]
            )
        assert e.value.unknown == [ThreeDAssetMapping._load({"assetId": 1, "nodeId": 1})]
